from daml._internal.metrics.ber import BER
from daml._internal.metrics.divergence import Divergence
from daml._internal.metrics.parity import Parity
from daml._internal.metrics.stats import ChannelStats, ImageStats
from daml._internal.metrics.uap import UAP

__all__ = ["BER", "Divergence", "Parity", "UAP", "ChannelStats", "ImageStats"]



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from functools import partial
from typing import Callable, Dict, Literal, Optional, Union

import numpy as np
from scipy.special import softmax
from scipy.stats import entropy

from .base import UpdateStrategy
from .ks import DriftKS
from .torch import get_device, preprocess_drift


def classifier_uncertainty(
    x: np.ndarray,
    model_fn: Callable,
    preds_type: Literal["probs", "logits"] = "probs",
) -> np.ndarray:
    """
    Evaluate model_fn on x and transform predictions to prediction uncertainties.

    Parameters
    ----------
    x
        Batch of instances.
    model_fn
        Function that evaluates a classification model on x in a single call (contains
        batching logic if necessary).
    preds_type
        Type of prediction output by the model. Options are 'probs' (in [0,1]) or
        'logits' (in [-inf,inf]).

    Returns
    -------
    A scalar indication of uncertainty of the model on each instance in x.
    """

    preds = model_fn(x)

    if preds_type == "probs":
        if np.abs(1 - np.sum(preds, axis=-1)).mean() > 1e-6:
            raise ValueError("Probabilities across labels should sum to 1")
        probs = preds
    elif preds_type == "logits":
        probs = softmax(preds, axis=-1)
    else:
        raise NotImplementedError("Only prediction types 'probs' and 'logits' supported.")

    uncertainties = entropy(probs, axis=-1)
    return uncertainties[:, None]  # Detectors expect N x d  # type: ignore


class DriftUncertainty:
    """
    Test for a change in the number of instances falling into regions on which the
    model is uncertain. Performs a K-S test on prediction entropies.

    Parameters
    ----------
    x_ref : np.ndarray
        Data used as reference distribution. Should be disjoint from the data the
        model was trained on for accurate p-values.
    model : Callable
        Classification model outputting class probabilities (or logits)
    p_val : float, default 0.05
        p-value used for the significance of the test.
    x_ref_preprocessed : bool, default False
        Whether the given reference data `x_ref` has been preprocessed yet. If
        `x_ref_preprocessed=True`, only the test data `x` will be preprocessed at
        prediction time. If `x_ref_preprocessed=False`, the reference data will
        also be preprocessed.
    update_x_ref : Optional[UpdateStrategy], default None
        Reference data can optionally be updated using an UpdateStrategy class. Update
        using the last n instances seen by the detector with
        :py:class:`daml.detectors.LastSeenUpdateStrategy`
        or via reservoir sampling with
        :py:class:`daml.detectors.ReservoirSamplingUpdateStrategy`.
    preds_type : Literal["probs", "logits"], default "logits"
        Type of prediction output by the model. Options are 'probs' (in [0,1]) or
        'logits' (in [-inf,inf]).
    batch_size : int, default 32
        Batch size used to evaluate model. Only relevant when backend has been
        specified for batch prediction.
    preprocess_batch_fn : Optional[Callable], default None
        Optional batch preprocessing function. For example to convert a list of
        objects to a batch which can be processed by the model.
    device : Optional[str], default None
        Device type used. The default None tries to use the GPU and falls back on
        CPU if needed. Can be specified by passing either 'cuda', 'gpu' or 'cpu'.
    input_shape : Optional[tuple], default None
        Shape of input data.
    """

    def __init__(
        self,
        x_ref: np.ndarray,
        model: Callable,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preds_type: Literal["probs", "logits"] = "probs",
        batch_size: int = 32,
        preprocess_batch_fn: Optional[Callable] = None,
        device: Optional[str] = None,
    ) -> None:
        def model_fn(x: np.ndarray) -> np.ndarray:
            return preprocess_drift(
                x,
                model,  # type: ignore
                batch_size=batch_size,
                preprocess_batch_fn=preprocess_batch_fn,
                device=get_device(device),
            )

        preprocess_fn = partial(
            classifier_uncertainty,
            model_fn=model_fn,
            preds_type=preds_type,
        )

        self._detector = DriftKS(
            x_ref=x_ref,
            p_val=p_val,
            x_ref_preprocessed=x_ref_preprocessed,
            update_x_ref=update_x_ref,
            preprocess_fn=preprocess_fn,
        )

    def predict(self, x: np.ndarray) -> Dict[str, Union[int, float, np.ndarray]]:
        """
        Predict whether a batch of data has drifted from the reference data.

        Parameters
        ----------
        x
            Batch of instances.

        Returns
        -------
        Dictionary containing the drift prediction, p-value, and threshold statistics.
        """
        return self._detector.predict(x)



from typing import Dict, List, Literal

import numpy as np

from daml._internal.flags import ImageHash
from daml._internal.metrics.stats import ImageStats


class Duplicates:
    """
    Finds the duplicate images in a dataset using xxhash for exact duplicates
    and pchash for near duplicates
    """

    def __init__(
        self,
        images: np.ndarray,
    ):
        self.stats = ImageStats(ImageHash.ALL)
        self.images = images

    def _get_duplicates(self) -> dict:
        exact = {}
        near = {}
        for i, value in enumerate(self.results["xxhash"]):
            exact.setdefault(value, []).append(i)
        for i, value in enumerate(self.results["pchash"]):
            near.setdefault(value, []).append(i)
        exact = [v for v in exact.values() if len(v) > 1]
        near = [v for v in near.values() if len(v) > 1 and not any(set(v).issubset(x) for x in exact)]

        return {
            "exact": sorted(exact),
            "near": sorted(near),
        }

    def evaluate(self) -> Dict[Literal["exact", "near"], List[int]]:
        """
        Returns duplicate image indices for both exact matches and near matches
    
        Returns
        -------
        Dict[Literal["exact", "near"], List[int]]
            Dictionary of exact and near match indices
        """
        self.stats.reset()
        self.stats.update(self.images)
        self.results = self.stats.compute()
        return self._get_duplicates()



from importlib.util import find_spec

__all__ = []

if find_spec("tensorflow") is not None:  # pragma: no cover
    from . import tensorflow

    __all__ += ["tensorflow"]

if find_spec("torch") is not None:  # pragma: no cover
    from . import torch

    __all__ += ["torch"]

del find_spec



import warnings
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
from matplotlib.figure import Figure
from scipy.optimize import basinhopping
from torch.utils.data import Dataset

from daml._internal.metrics.base import EvaluateMixin

STEPS_KEY = "_STEPS_"
PARAMS_KEY = "_CURVE_PARAMS_"

SufficiencyOutput = Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]


def f_out(n_i: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Calculates the line of best fit based on its free parameters

    Parameters
    ----------
    n_i : np.ndarray
        Array of sample sizes
    x : np.ndarray
        Array of inverse power curve coefficients

    Returns
    -------
    np.ndarray
        Data points for the line of best fit
    """
    return x[0] * n_i ** (-x[1]) + x[2]


def f_inv_out(y_i: np.ndarray, x: np.ndarray) -> np.ndarray:
    """
    Inverse function for f_out()

    Parameters
    ----------
    y_i : np.ndarray
        Data points for the line of best fit
    x : np.ndarray
        Array of inverse power curve coefficients

    Returns
    -------
    np.ndarray
        Array of sample sizes
    """
    n_i = ((y_i - x[2]) / x[0]) ** (-1 / x[1])
    return n_i


def calc_params(p_i: np.ndarray, n_i: np.ndarray, niter: int) -> np.ndarray:
    """
    Retrieves the inverse power curve coefficients for the line of best fit.
    Global minimization is done via basin hopping. More info on this algorithm
    can be found here: https://arxiv.org/abs/cond-mat/9803344 .

    Parameters
    ----------
    p_i : np.ndarray
        Array of corresponding losses
    n_i : np.ndarray
        Array of sample sizes
    niter : int
        Number of iterations to perform in the basin-hopping
        numerical process to curve-fit p_i

    Returns
    -------
    np.ndarray
        Array of parameters to recreate line of best fit
    """

    def is_valid(f_new, x_new, f_old, x_old):
        return f_new != np.nan

    def f(x):
        try:
            return np.sum(np.square(p_i - f_out(n_i, x)))
        except RuntimeWarning:
            return np.nan

    with warnings.catch_warnings():
        warnings.filterwarnings("error", category=RuntimeWarning)
        res = basinhopping(
            f,
            np.array([0.5, 0.5, 0.1]),
            niter=niter,
            stepsize=1.0,
            minimizer_kwargs={"method": "Powell"},
            accept_test=is_valid,
            niter_success=200,
        )
    return res.x


def reset_parameters(model: nn.Module):
    """
    Re-initializes each layer in the model using
    the layer's defined weight_init function
    """

    @torch.no_grad()
    def weight_reset(m: nn.Module):
        # Check if the current module has reset_parameters
        reset_parameters = getattr(m, "reset_parameters", None)
        if callable(reset_parameters):
            m.reset_parameters()  # type: ignore

    # Applies fn recursively to every submodule see:
    # https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    return model.apply(fn=weight_reset)


def validate_dataset_len(dataset: Dataset) -> int:
    if not hasattr(dataset, "__len__"):
        raise TypeError("Must provide a dataset with a length attribute")
    length: int = dataset.__len__()  # type: ignore
    if length <= 0:
        raise ValueError("Dataset length must be greater than 0")
    return length


def validate_output(data: SufficiencyOutput):
    """Ensure the sufficiency data used is not malformed"""
    if not all(key in data for key in [STEPS_KEY, PARAMS_KEY]):
        raise KeyError(f"{STEPS_KEY} and {PARAMS_KEY} are required keys for Sufficiency output.")
    c = len(data[STEPS_KEY])
    for m, v in data.items():
        if m in [STEPS_KEY, PARAMS_KEY]:
            continue
        v = cast(np.ndarray, v)
        c_v = v.shape[1] if v.ndim > 1 else len(v)
        if c != c_v:
            raise ValueError("f{m} does not contain the expected number ({c}) of data points.")


def project_steps(params: np.ndarray, projection: np.ndarray) -> np.ndarray:
    """Projects the measures for each value of X

    Parameters
    ----------
    params : np.ndarray
        Inverse power curve coefficients used to calculate projection
    projection : np.ndarray
        Steps to extrapolate

    Returns
    -------
    np.ndarray
        Extrapolated measure values at each projection step

    """
    return 1 - f_out(projection, params)


def inv_project_steps(params: np.ndarray, targets: np.ndarray) -> np.ndarray:
    """Inverse function for project_steps()

    Parameters
    ----------
    params : np.ndarray
        Inverse power curve coefficients used to calculate projection
    targets : np.ndarray
        Desired measure values

    Returns
    -------
    np.ndarray
        Array of sample sizes, or 0 if overflow
    """
    steps = f_inv_out(1 - np.array(targets), params)
    steps[np.isnan(steps)] = 0
    return np.ceil(steps).astype(np.int64)


def get_curve_params(measures: Dict[str, np.ndarray], ranges: np.ndarray, niter: int) -> Dict[str, np.ndarray]:
    """Calculates and aggregates parameters for both single and multi-class metrics"""
    output = {}
    for name, measure in measures.items():
        measure = cast(np.ndarray, measure)
        if measure.ndim > 1:
            result = []
            for value in measure:
                result.append(calc_params(1 - value, ranges, niter))
            output[name] = np.array(result)
        else:
            output[name] = calc_params(1 - measure, ranges, niter)
    return output


def plot_measure(
    name: str,
    steps: np.ndarray,
    measure: np.ndarray,
    params: np.ndarray,
    projection: np.ndarray,
) -> Figure:
    fig = plt.figure()
    fig = cast(Figure, fig)
    fig.tight_layout()

    ax = fig.add_subplot(111)

    ax.set_title(f"{name} Sufficiency")
    ax.set_ylabel(f"{name}")
    ax.set_xlabel("Steps")

    # Plot measure over each step
    ax.scatter(steps, measure, label=f"Model Results ({name})", s=15, c="black")

    # Plot extrapolation
    ax.plot(
        projection,
        project_steps(params, projection),
        linestyle="dashed",
        label=f"Potential Model Results ({name})",
    )

    ax.legend()
    return fig


class Sufficiency(EvaluateMixin):
    """
    Project dataset sufficiency using given a model and evaluation criteria

    Parameters
    ----------
    model : nn.Module
        Model that will be trained for each subset of data
    train_ds : Dataset
        Full training data that will be split for each run
    test_ds : Dataset
        Data that will be used for every run's evaluation
    train_fn : Callable[[nn.Module, Dataset, Sequence[int]], None]
        Function which takes a model (torch.nn.Module), a dataset
        (torch.utils.data.Dataset), indices to train on and executes model
        training against the data.
    eval_fn : Callable[[nn.Module, Dataset], Dict[str, float]]
        Function which takes a model (torch.nn.Module), a dataset
        (torch.utils.data.Dataset) and returns a dictionary of metric
        values (Dict[str, float]) which is used to assess model performance
        given the model and data.
    runs : int, default 1
        Number of models to run over all subsets
    substeps : int, default 5
        Total number of dataset partitions that each model will train on
    train_kwargs : Dict[str, Any] | None, default None
        Additional arguments required for custom training function
    eval_kwargs : Dict[str, Any] | None, default None
        Additional arguments required for custom evaluation function
    """

    def __init__(
        self,
        model: nn.Module,
        train_ds: Dataset,
        test_ds: Dataset,
        train_fn: Callable[[nn.Module, Dataset, Sequence[int]], None],
        eval_fn: Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]],
        runs: int = 1,
        substeps: int = 5,
        train_kwargs: Optional[Dict[str, Any]] = None,
        eval_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.model = model
        self.train_ds = train_ds
        self.test_ds = test_ds
        self.train_fn = train_fn
        self.eval_fn = eval_fn
        self.runs = runs
        self.substeps = substeps
        self.train_kwargs = train_kwargs
        self.eval_kwargs = eval_kwargs

    @property
    def train_ds(self):
        return self._train_ds

    @train_ds.setter
    def train_ds(self, value: Dataset):
        self._train_ds = value
        self._length = validate_dataset_len(value)

    @property
    def test_ds(self):
        return self._test_ds

    @test_ds.setter
    def test_ds(self, value: Dataset):
        validate_dataset_len(value)
        self._test_ds = value

    @property
    def train_fn(self) -> Callable[[nn.Module, Dataset, Sequence[int]], None]:
        return self._train_fn

    @train_fn.setter
    def train_fn(self, value: Callable[[nn.Module, Dataset, Sequence[int]], None]):
        if not callable(value):
            raise TypeError("Must provide a callable for train_fn.")
        self._train_fn = value

    @property
    def eval_fn(
        self,
    ) -> Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]]:
        return self._eval_fn

    @eval_fn.setter
    def eval_fn(
        self,
        value: Callable[[nn.Module, Dataset], Union[Dict[str, float], Dict[str, np.ndarray]]],
    ):
        if not callable(value):
            raise TypeError("Must provide a callable for eval_fn.")
        self._eval_fn = value

    @property
    def train_kwargs(self) -> Dict[str, Any]:
        return self._train_kwargs

    @train_kwargs.setter
    def train_kwargs(self, value: Optional[Dict[str, Any]]):
        self._train_kwargs = {} if value is None else value

    @property
    def eval_kwargs(self) -> Dict[str, Any]:
        return self._eval_kwargs

    @eval_kwargs.setter
    def eval_kwargs(self, value: Optional[Dict[str, Any]]):
        self._eval_kwargs = {} if value is None else value

    def evaluate(self, eval_at: Optional[np.ndarray] = None, niter: int = 1000) -> SufficiencyOutput:
        """
        Creates data indices, trains models, and returns plotting data

        Parameters
        ----------
        eval_at : Optional[np.ndarray]
            Specify this to collect accuracies over a specific set of dataset lengths, rather
            than letting Sufficiency internally create the lengths to evaluate at.
        niter : int, default 1000
            Iterations to perform when using the basin-hopping method to curve-fit measure(s).

        Returns
        -------
        Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]
            Dictionary containing the average of each measure per substep
        """
        if eval_at is not None:
            ranges = eval_at
        else:
            geomshape = (
                0.01 * self._length,
                self._length,
                self.substeps,
            )  # Start, Stop, Num steps
            ranges = np.geomspace(*geomshape).astype(np.int64)
        substeps = len(ranges)
        measures = {}

        # Run each model over all indices
        for _ in range(self.runs):
            # Create a randomized set of indices to use
            indices = np.random.randint(0, self._length, size=self._length)
            # Reset the network weights to "create" an untrained model
            model = reset_parameters(self.model)
            # Run the model with each substep of data
            for iteration, substep in enumerate(ranges):
                # train on subset of train data
                self.train_fn(
                    model,
                    self.train_ds,
                    indices[:substep].tolist(),
                    **self.train_kwargs,
                )

                # evaluate on test data
                measure = self.eval_fn(model, self.test_ds, **self.eval_kwargs)

                # Keep track of each measures values
                for name, value in measure.items():
                    if name in [STEPS_KEY, PARAMS_KEY]:
                        raise KeyError(f"Cannot use reserved name '{name}' as a metric name.")

                    # Sum result into current substep iteration to be averaged later
                    value = np.array(value).ravel()
                    if name not in measures:
                        measures[name] = np.zeros(substeps if len(value) == 1 else (substeps, len(value)))
                    measures[name][iteration] += value

        # The mean for each measure must be calculated before being returned
        measures = {k: (v / self.runs).T for k, v in measures.items()}
        params_output = get_curve_params(measures, ranges, niter)
        output = {STEPS_KEY: ranges, PARAMS_KEY: params_output}
        output.update(measures)
        return output

    @classmethod
    def project(
        cls,
        data: SufficiencyOutput,
        projection: Union[int, Sequence[int], np.ndarray],
    ) -> Dict[str, np.ndarray]:
        """Projects the measures for each value of X

        Parameters
        ----------
        data : Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]
            Dataclass containing the average of each measure per substep
        steps : Union[int, np.ndarray]
            Step or steps to project
        niter : int, default 200
            Number of iterations to perform in the basin-hopping
            numerical process to curve-fit data

        Raises
        ------
        KeyError
            If STEPS_KEY or measure is not a valid key
        ValueError
            If the length of data points in the measures do not match
            If the steps are not int, Sequence[int] or an ndarray
        """
        validate_output(data)
        projection = [projection] if isinstance(projection, int) else projection
        projection = np.array(projection) if isinstance(projection, Sequence) else projection
        if not isinstance(projection, np.ndarray):
            raise ValueError("'steps' must be an int, Sequence[int] or ndarray")

        output = {}
        output[STEPS_KEY] = projection
        for name, measures in data.items():
            if name in [STEPS_KEY, PARAMS_KEY]:
                continue
            measures = cast(np.ndarray, measures)
            if measures.ndim > 1:
                result = []
                for i in range(len(measures)):
                    projected = project_steps(data[PARAMS_KEY][name][i], projection)
                    result.append(projected)
                output[name] = np.array(result).T
            else:
                output[name] = project_steps(data[PARAMS_KEY][name], projection)
        return output

    @classmethod
    def plot(cls, data: SufficiencyOutput, class_names: Optional[Sequence[str]] = None) -> List[Figure]:
        """Plotting function for data sufficiency tasks

        Parameters
        ----------
        data : Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]
            Dataclass containing the average of each measure per substep

        Returns
        -------
        List[plt.Figure]
            List of Figures for each measure

        Raises
        ------
        KeyError
            If STEPS_KEY or measure is not a valid key
        ValueError
            If the length of data points in the measures do not match
        """
        validate_output(data)

        # X, y data
        steps = cast(np.ndarray, data[STEPS_KEY])

        # Extrapolation parameters
        last_X = steps[-1]
        geomshape = (0.01 * last_X, last_X * 4, len(steps))
        extrapolated = np.geomspace(*geomshape).astype(np.int64)

        # Stores all plots
        plots = []

        # Create a plot for each measure on one figure
        for name, measures in data.items():
            if name in [STEPS_KEY, PARAMS_KEY]:
                continue
            measures = cast(np.ndarray, measures)
            if measures.ndim > 1:
                if class_names is not None and len(measures) != len(class_names):
                    raise IndexError("Class name count does not align with measures")
                for i, measure in enumerate(measures):
                    class_name = str(i) if class_names is None else class_names[i]
                    fig = plot_measure(
                        f"{name}_{class_name}",
                        steps,
                        measure,
                        data[PARAMS_KEY][name][i],
                        extrapolated,
                    )
                    plots.append(fig)

            else:
                fig = plot_measure(name, steps, measures, data[PARAMS_KEY][name], extrapolated)
                plots.append(fig)

        return plots

    @classmethod
    def inv_project(cls, targets: Dict[str, np.ndarray], data: SufficiencyOutput) -> Dict[str, np.ndarray]:
        """
        Calculate training samples needed to achieve target model metric values.

        Parameters
        ----------
        targets : Dict[str, np.ndarray]
            Dictionary of target metric scores (from 0.0 to 1.0) that we want
            to achieve, where the key is the name of the metric.

        data : Dict[str, Union[np.ndarray, Dict[str, np.ndarray]]]
            Dataclass containing the average of each measure per substep

        Returns
        -------
        Dict[str, np.ndarray]
            List of the number of training samples needed to achieve each
            corresponding entry in targets
        """

        validate_output(data)

        projection = {}

        for name, target in targets.items():
            if name not in data:
                continue

            measure = cast(np.ndarray, data[name])
            if measure.ndim > 1:
                projection[name] = np.zeros((len(measure), len(target)))
                for i in range(len(measure)):
                    projection[name][i] = inv_project_steps(
                        data[PARAMS_KEY][name][i], target[i] if target.ndim == measure.ndim else target
                    )
            else:
                projection[name] = inv_project_steps(data[PARAMS_KEY][name], target)

        return projection



from importlib.util import find_spec

if find_spec("torch") is not None:  # pragma: no cover
    from daml._internal.workflows.sufficiency import Sufficiency

    __all__ = ["Sufficiency"]

del find_spec



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable, Iterable, Optional, Tuple, cast

import keras.api._v2.keras as keras
import numpy as np
import tensorflow as tf


def trainer(
    model: keras.Model,
    x_train: np.ndarray,
    y_train: Optional[np.ndarray] = None,
    loss_fn: Optional[Callable[..., tf.Tensor]] = None,
    optimizer: keras.optimizers.Optimizer = keras.optimizers.Adam,
    preprocess_fn: Optional[Callable[[tf.Tensor], tf.Tensor]] = None,
    epochs: int = 20,
    reg_loss_fn: Callable[[keras.Model], tf.Tensor] = (lambda _: cast(tf.Tensor, tf.Variable(0, dtype=tf.float32))),
    batch_size: int = 64,
    buffer_size: int = 1024,
    verbose: bool = True,
) -> None:
    """
    Train TensorFlow model.

    Parameters
    ----------
    model
        Model to train.
    loss_fn
        Loss function used for training.
    x_train
        Training data.
    y_train
        Training labels.
    optimizer
        Optimizer used for training.
    preprocess_fn
        Preprocessing function applied to each training batch.
    epochs
        Number of training epochs.
    reg_loss_fn
        Allows an additional regularisation term to be defined as reg_loss_fn(model)
    batch_size
        Batch size used for training.
    buffer_size
        Maximum number of elements that will be buffered when prefetching.
    verbose
        Whether to print training progress.
    """
    loss_fn = loss_fn() if isinstance(loss_fn, type) else loss_fn
    optimizer = optimizer() if isinstance(optimizer, type) else optimizer

    train_data = x_train if y_train is None else (x_train, y_train)
    dataset = tf.data.Dataset.from_tensor_slices(train_data)
    dataset = dataset.shuffle(buffer_size=buffer_size).batch(batch_size)
    n_minibatch = len(dataset)

    # iterate over epochs
    for epoch in range(epochs):
        pbar = keras.utils.Progbar(n_minibatch, 1) if verbose else None
        if hasattr(dataset, "on_epoch_end"):
            dataset.on_epoch_end()  # type: ignore py39
        loss_val_ma = 0.0
        for step, data in enumerate(dataset):
            x, y = cast(Tuple[tf.Tensor, Optional[tf.Tensor]], data if isinstance(data, tuple) else (data, None))
            if isinstance(preprocess_fn, Callable):
                x = preprocess_fn(x)
            with tf.GradientTape() as tape:
                y_hat = model(x)
                y = x if y is None else y
                if isinstance(loss_fn, Callable):
                    args = [y] + list(y_hat) if isinstance(y_hat, Tuple) else [y, y_hat]
                    loss = loss_fn(*args)
                else:
                    loss = cast(tf.Tensor, tf.constant(0.0, dtype=tf.float32))
                if model.losses:  # additional model losses
                    loss = cast(tf.Tensor, tf.add(sum(model.losses), loss))
                loss = cast(tf.Tensor, tf.add(reg_loss_fn(model), loss))  # alternative way they might be specified

            grads = cast(Iterable, tape.gradient(loss, model.trainable_weights))
            optimizer.apply_gradients(zip(grads, model.trainable_weights))
            if pbar is not None:
                loss_val = getattr(loss, "numpy")() if hasattr(loss, "numpy") else np.float32(0.0)
                if loss_val.shape and loss_val.shape[0] != batch_size:
                    if len(loss_val.shape) == 1:
                        shape = (batch_size - loss_val.shape[0],)
                    elif len(loss_val.shape) == 2:
                        shape = (batch_size - loss_val.shape[0], loss_val.shape[1])
                    else:
                        continue
                    add_mean = np.ones(shape) * loss_val.mean()
                    loss_val = np.r_[loss_val, add_mean]
                loss_val_ma = loss_val_ma + (loss_val - loss_val_ma) / (step + 1)
                pbar_values = [("loss_ma", loss_val_ma)]
                pbar.add(1, values=pbar_values)



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable, Dict, Optional, Tuple, Union

import numpy as np
import torch

from .base import BaseDrift, UpdateStrategy, preprocess_x, update_x_ref
from .torch import GaussianRBF, get_device, mmd2_from_kernel_matrix


class DriftMMD(BaseDrift):
    """
    Maximum Mean Discrepancy (MMD) data drift detector using a permutation test.

    Parameters
    ----------
    x_ref : np.ndarray
        Data used as reference distribution.
    p_val : float, default 0.05
        p-value used for the significance of the permutation test.
    x_ref_preprocessed : bool, default False
        Whether the given reference data `x_ref` has been preprocessed yet. If
        `x_ref_preprocessed=True`, only the test data `x` will be preprocessed
        at prediction time. If `x_ref_preprocessed=False`, the reference data
        will also be preprocessed.
    preprocess_at_init : bool, default True
        Whether to preprocess the reference data when the detector is instantiated.
        Otherwise, the reference data will be preprocessed at prediction time. Only
        applies if `x_ref_preprocessed=False`.
    update_x_ref : Optional[UpdateStrategy], default None
        Reference data can optionally be updated using an UpdateStrategy class. Update
        using the last n instances seen by the detector with
        :py:class:`daml.detectors.LastSeenUpdateStrategy`
        or via reservoir sampling with
        :py:class:`daml.detectors.ReservoirSamplingUpdateStrategy`.
    preprocess_fn : Optional[Callable], default None
        Function to preprocess the data before computing the data drift metrics.
    kernel : Callable, default :py:class:`daml.detectors.GaussianRBF`
        Kernel used for the MMD computation, defaults to Gaussian RBF kernel.
    sigma : Optional[np.ndarray], default None
        Optionally set the GaussianRBF kernel bandwidth. Can also pass multiple
        bandwidth values as an array. The kernel evaluation is then averaged over
        those bandwidths.
    configure_kernel_from_x_ref : bool, default True
        Whether to already configure the kernel bandwidth from the reference data.
    n_permutations : int, default 100
        Number of permutations used in the permutation test.
    device : Optional[str], default None
        Device type used. The default None uses the GPU and falls back on CPU.
        Can be specified by passing either 'cuda', 'gpu' or 'cpu'.
    """

    def __init__(
        self,
        x_ref: np.ndarray,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preprocess_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        kernel: Callable = GaussianRBF,
        sigma: Optional[np.ndarray] = None,
        configure_kernel_from_x_ref: bool = True,
        n_permutations: int = 100,
        device: Optional[str] = None,
    ) -> None:
        super().__init__(x_ref, p_val, x_ref_preprocessed, update_x_ref, preprocess_fn)

        self.infer_sigma = configure_kernel_from_x_ref
        if configure_kernel_from_x_ref and isinstance(sigma, np.ndarray):
            self.infer_sigma = False

        self.n_permutations = n_permutations  # nb of iterations through permutation test

        # set device
        self.device = get_device(device)

        # initialize kernel
        sigma_tensor = torch.from_numpy(sigma).to(self.device) if isinstance(sigma, np.ndarray) else None
        self.kernel = kernel(sigma_tensor).to(self.device) if kernel == GaussianRBF else kernel

        # compute kernel matrix for the reference data
        if self.infer_sigma or isinstance(sigma_tensor, torch.Tensor):
            x = torch.from_numpy(self.x_ref).to(self.device)
            self.k_xx = self.kernel(x, x, infer_sigma=self.infer_sigma)
            self.infer_sigma = False
        else:
            self.k_xx, self.infer_sigma = None, True

    def _kernel_matrix(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Compute and return full kernel matrix between arrays x and y."""
        k_xy = self.kernel(x, y, self.infer_sigma)
        k_xx = self.k_xx if self.k_xx is not None and self.update_x_ref is None else self.kernel(x, x)
        k_yy = self.kernel(y, y)
        kernel_mat = torch.cat([torch.cat([k_xx, k_xy], 1), torch.cat([k_xy.T, k_yy], 1)], 0)
        return kernel_mat

    @preprocess_x
    def score(self, x: np.ndarray) -> Tuple[float, float, float]:
        """
        Compute the p-value resulting from a permutation test using the maximum mean
        discrepancy as a distance measure between the reference data and the data to
        be tested.

        Parameters
        ----------
        x
            Batch of instances.

        Returns
        -------
        p-value obtained from the permutation test, the MMD^2 between the reference and
        test set, and the MMD^2 threshold above which drift is flagged.
        """
        x_ref = torch.from_numpy(self.x_ref).to(self.device)
        n = x.shape[0]
        kernel_mat = self._kernel_matrix(x_ref, torch.from_numpy(x).to(self.device))
        kernel_mat = kernel_mat - torch.diag(kernel_mat.diag())  # zero diagonal
        mmd2 = mmd2_from_kernel_matrix(kernel_mat, n, permute=False, zero_diag=False)
        mmd2_permuted = torch.Tensor(
            [mmd2_from_kernel_matrix(kernel_mat, n, permute=True, zero_diag=False) for _ in range(self.n_permutations)]
        )
        mmd2, mmd2_permuted = mmd2.cpu(), mmd2_permuted.cpu()
        p_val = (mmd2 <= mmd2_permuted).float().mean()
        # compute distance threshold
        idx_threshold = int(self.p_val * len(mmd2_permuted))
        distance_threshold = torch.sort(mmd2_permuted, descending=True).values[idx_threshold]
        return p_val.numpy().item(), mmd2.numpy().item(), distance_threshold.numpy()

    @preprocess_x
    @update_x_ref
    def predict(
        self,
        x: np.ndarray,
    ) -> Dict[str, Union[int, float]]:
        """
        Predict whether a batch of data has drifted from the reference data and then
        updates reference data using specified strategy.

        Parameters
        ----------
        x
            Batch of instances.

        Returns
        -------
        Dictionary containing the drift prediction, p-value, threshold and MMD metric.
        """
        # compute drift scores
        p_val, dist, distance_threshold = self.score(x)
        drift_pred = int(p_val < self.p_val)

        # populate drift dict
        return {
            "is_drift": drift_pred,
            "p_val": p_val,
            "threshold": self.p_val,
            "distance": dist,
            "distance_threshold": distance_threshold,
        }



from enum import Flag, auto
from typing import Union


class auto_all:
    def __get__(self, _, cls):
        return ~cls(0)


class ImageHash(Flag):
    XXHASH = auto()
    PCHASH = auto()
    ALL = auto_all()


class ImageProperty(Flag):
    WIDTH = auto()
    HEIGHT = auto()
    SIZE = auto()
    ASPECT_RATIO = auto()
    CHANNELS = auto()
    DEPTH = auto()
    ALL = auto_all()


class ImageVisuals(Flag):
    BRIGHTNESS = auto()
    BLURRINESS = auto()
    MISSING = auto()
    ZERO = auto()
    ALL = auto_all()


class ImageStatistics(Flag):
    MEAN = auto()
    STD = auto()
    VAR = auto()
    SKEW = auto()
    KURTOSIS = auto()
    ENTROPY = auto()
    PERCENTILES = auto()
    HISTOGRAM = auto()
    ALL = auto_all()


ImageStatsFlags = Union[ImageHash, ImageProperty, ImageVisuals, ImageStatistics]
LinterFlags = Union[ImageProperty, ImageVisuals, ImageStatistics]



# type: ignore

"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

import functools
import warnings
from typing import Optional

import keras.api._v2.keras as keras
import numpy as np
import tensorflow as tf
from tensorflow_probability.python.bijectors import bijector
from tensorflow_probability.python.distributions import (
    categorical,
    distribution,
    independent,
    logistic,
    mixture_same_family,
    quantized_distribution,
    transformed_distribution,
)
from tensorflow_probability.python.internal import (
    dtype_util,
    prefer_static,
    reparameterization,
    tensor_util,
    tensorshape_util,
)

__all__ = [
    "Shift",
]


class WeightNorm(keras.layers.Wrapper):
    def __init__(self, layer, data_init: bool = True, **kwargs):
        """Layer wrapper to decouple magnitude and direction of the layer's weights.

        This wrapper reparameterizes a layer by decoupling the weight's
        magnitude and direction. This speeds up convergence by improving the
        conditioning of the optimization problem. It has an optional data-dependent
        initialization scheme, in which initial values of weights are set as functions
        of the first minibatch of data. Both the weight normalization and data-
        dependent initialization are described in [Salimans and Kingma (2016)][1].

        Parameters
        ----------
        layer
            A `keras.layers.Layer` instance. Supported layer types are
            `Dense`, `Conv2D`, and `Conv2DTranspose`. Layers with multiple inputs
            are not supported.
        data_init
            If `True` use data dependent variable initialization.
        **kwargs
            Additional keyword args passed to `keras.layers.Wrapper`.

        Raises
        ------
        ValueError
            If `layer` is not a `keras.layers.Layer` instance.
        """
        if not isinstance(layer, keras.layers.Layer):
            raise ValueError(
                "Please initialize `WeightNorm` layer with a `keras.layers.Layer` " f"instance. You passed: {layer}"
            )

        layer_type = type(layer).__name__
        if layer_type not in ["Dense", "Conv2D", "Conv2DTranspose"]:
            warnings.warn(
                "`WeightNorm` is tested only for `Dense`, `Conv2D`, and "
                f"`Conv2DTranspose` layers. You passed a layer of type `{layer_type}`"
            )

        super().__init__(layer, **kwargs)

        self.data_init = data_init
        self._track_trackable(layer, name="layer")
        self.filter_axis = -2 if layer_type == "Conv2DTranspose" else -1

    def _compute_weights(self):
        """Generate weights with normalization."""
        # Determine the axis along which to expand `g` so that `g` broadcasts to
        # the shape of `v`.
        new_axis = -self.filter_axis - 3

        self.layer.kernel = tf.nn.l2_normalize(self.v, axis=self.kernel_norm_axes) * tf.expand_dims(self.g, new_axis)

    def _init_norm(self):
        """Set the norm of the weight vector."""
        kernel_norm = tf.sqrt(tf.reduce_sum(tf.square(self.v), axis=self.kernel_norm_axes))
        self.g.assign(kernel_norm)

    def _data_dep_init(self, inputs):
        """Data dependent initialization."""
        # Normalize kernel first so that calling the layer calculates
        # `tf.dot(v, x)/tf.norm(v)` as in (5) in ([Salimans and Kingma, 2016][1]).
        self._compute_weights()

        activation = self.layer.activation
        self.layer.activation = None

        use_bias = self.layer.bias is not None
        if use_bias:
            bias = self.layer.bias
            self.layer.bias = tf.zeros_like(bias)

        # Since the bias is initialized as zero, setting the activation to zero and
        # calling the initialized layer (with normalized kernel) yields the correct
        # computation ((5) in Salimans and Kingma (2016))
        x_init = self.layer(inputs)
        norm_axes_out = list(range(x_init.shape.rank - 1))
        m_init, v_init = tf.nn.moments(x_init, norm_axes_out)
        scale_init = 1.0 / tf.sqrt(v_init + 1e-10)

        self.g.assign(self.g * scale_init)
        if use_bias:
            self.layer.bias = bias
            self.layer.bias.assign(-m_init * scale_init)
        self.layer.activation = activation

    def build(self, input_shape=None):
        """Build `Layer`.

        Parameters
        ----------
        input_shape
            The shape of the input to `self.layer`.

        Raises
        ------
        ValueError
            If `Layer` does not contain a `kernel` of weights.
        """
        input_shape = tf.TensorShape(input_shape).as_list()
        input_shape[0] = None
        self.input_spec = keras.layers.InputSpec(shape=input_shape)

        if not self.layer.built:
            self.layer.build(input_shape)

            if not hasattr(self.layer, "kernel"):
                raise ValueError("`WeightNorm` must wrap a layer that contains a `kernel` for weights")

            self.kernel_norm_axes = list(range(self.layer.kernel.shape.ndims))
            self.kernel_norm_axes.pop(self.filter_axis)

            self.v = self.layer.kernel

            # to avoid a duplicate `kernel` variable after `build` is called
            self.layer.kernel = None
            self.g = self.add_weight(
                name="g",
                shape=(int(self.v.shape[self.filter_axis]),),
                initializer="ones",
                dtype=self.v.dtype,
                trainable=True,
            )
            self.initialized = self.add_weight(name="initialized", dtype=tf.bool, trainable=False)
            self.initialized.assign(False)

        super().build()

    @tf.function
    def call(self, inputs):
        """Call `Layer`."""
        if not self.initialized:
            if self.data_init:
                self._data_dep_init(inputs)
            else:  # initialize `g` as the norm of the initialized kernel
                self._init_norm()

            self.initialized.assign(True)

        self._compute_weights()
        output = self.layer(inputs)
        return output

    def compute_output_shape(self, input_shape):
        return tf.TensorShape(self.layer.compute_output_shape(input_shape).as_list())


class Shift(bijector.Bijector):
    def __init__(self, shift, validate_args=False, name="shift"):
        """Instantiates the `Shift` bijector which computes `Y = g(X; shift) = X + shift`
        where `shift` is a numeric `Tensor`.

        Parameters
        ----------
        shift
            Floating-point `Tensor`.
        validate_args
            Python `bool` indicating whether arguments should be checked for correctness.
        name
            Python `str` name given to ops managed by this object.
        """
        with tf.name_scope(name) as name:
            dtype = dtype_util.common_dtype([shift], dtype_hint=tf.float32)
            self._shift = tensor_util.convert_nonref_to_tensor(shift, dtype=dtype, name="shift")
            super().__init__(
                forward_min_event_ndims=0,
                is_constant_jacobian=True,
                dtype=dtype,
                validate_args=validate_args,
                name=name,
            )

    @property
    def shift(self):
        """The `shift` `Tensor` in `Y = X + shift`."""
        return self._shift

    @classmethod
    def _is_increasing(cls):
        return True

    def _forward(self, x):
        return x + self.shift

    def _inverse(self, y):
        return y - self.shift

    def _forward_log_det_jacobian(self, x):
        # is_constant_jacobian = True for this bijector, hence the
        # `log_det_jacobian` need only be specified for a single input, as this will
        # be tiled to match `event_ndims`.
        return tf.zeros([], dtype=dtype_util.base_dtype(x.dtype))


class PixelCNN(distribution.Distribution):
    """
    Construct Pixel CNN++ distribution.

    Parameters
    ----------
    image_shape
        3D `TensorShape` or tuple for the `[height, width, channels]` dimensions of the image.
    conditional_shape
        `TensorShape` or tuple for the shape of the conditional input, or `None` if there is no conditional input.
    num_resnet
        The number of layers (shown in Figure 2 of [2]) within each highest-level block of Figure 2 of [1].
    num_hierarchies
        The number of highest-level blocks (separated by expansions/contractions of dimensions in Figure 2 of [1].)
    num_filters
        The number of convolutional filters.
    num_logistic_mix
        Number of components in the logistic mixture distribution.
    receptive_field_dims
        Height and width in pixels of the receptive field of the convolutional layers above and to the left
        of a given pixel. The width (second element of the tuple) should be odd. Figure 1 (middle) of [2]
        shows a receptive field of (3, 5) (the row containing the current pixel is included in the height).
        The default of (3, 3) was used to produce the results in [1].
    dropout_p
        The dropout probability. Should be between 0 and 1.
    resnet_activation
        The type of activation to use in the resnet blocks. May be 'concat_elu', 'elu', or 'relu'.
    l2_weight
        The L2 regularization weight.
    use_weight_norm
        If `True` then use weight normalization (works only in Eager mode).
    use_data_init
        If `True` then use data-dependent initialization (has no effect if `use_weight_norm` is `False`).
    high
        The maximum value of the input data (255 for an 8-bit image).
    low
        The minimum value of the input data.
    dtype
        Data type of the `Distribution`.
    name
        The name of the `Distribution`.
    """

    def __init__(
        self,
        image_shape: tuple,
        conditional_shape: Optional[tuple] = None,
        num_resnet: int = 5,
        num_hierarchies: int = 3,
        num_filters: int = 160,
        num_logistic_mix: int = 10,
        receptive_field_dims: tuple = (3, 3),
        dropout_p: float = 0.5,
        resnet_activation: str = "concat_elu",
        l2_weight: float = 0.0,
        use_weight_norm: bool = True,
        use_data_init: bool = True,
        high: int = 255,
        low: int = 0,
        dtype=tf.float32,
        name: str = "PixelCNN",
    ) -> None:
        parameters = dict(locals())
        with tf.name_scope(name) as name:
            super().__init__(
                dtype=dtype,
                reparameterization_type=reparameterization.NOT_REPARAMETERIZED,
                validate_args=False,
                allow_nan_stats=True,
                parameters=parameters,
                name=name,
            )

            if not tensorshape_util.is_fully_defined(image_shape):
                raise ValueError("`image_shape` must be fully defined.")

            if conditional_shape is not None and not tensorshape_util.is_fully_defined(conditional_shape):
                raise ValueError("`conditional_shape` must be fully defined.")

            if tensorshape_util.rank(image_shape) != 3:
                raise ValueError("`image_shape` must have length 3, representing [height, width, channels] dimensions.")

            self._high = tf.cast(high, self.dtype)
            self._low = tf.cast(low, self.dtype)
            self._num_logistic_mix = num_logistic_mix
            self.network = _PixelCNNNetwork(
                dropout_p=dropout_p,
                num_resnet=num_resnet,
                num_hierarchies=num_hierarchies,
                num_filters=num_filters,
                num_logistic_mix=num_logistic_mix,
                receptive_field_dims=receptive_field_dims,
                resnet_activation=resnet_activation,
                l2_weight=l2_weight,
                use_weight_norm=use_weight_norm,
                use_data_init=use_data_init,
                dtype=dtype,
            )

            image_input_shape = tensorshape_util.concatenate([None], image_shape)
            if conditional_shape is None:
                input_shape = image_input_shape
            else:
                conditional_input_shape = tensorshape_util.concatenate([None], conditional_shape)
                input_shape = [image_input_shape, conditional_input_shape]

            self.image_shape = image_shape
            self.conditional_shape = conditional_shape
            self.network.build(input_shape)

    def _make_mixture_dist(self, component_logits, locs, scales, return_per_feature: bool = False):
        """Builds a mixture of quantized logistic distributions.

        Parameters
        ----------
        component_logits
            4D `Tensor` of logits for the Categorical distribution
            over Quantized Logistic mixture components. Dimensions are `[batch_size,
            height, width, num_logistic_mix]`.
        locs
            4D `Tensor` of location parameters for the Quantized Logistic
            mixture components. Dimensions are `[batch_size, height, width,
            num_logistic_mix, num_channels]`.
        scales
            4D `Tensor` of location parameters for the Quantized Logistic
            mixture components. Dimensions are `[batch_size, height, width,
            num_logistic_mix, num_channels]`.
        return_per_feature
            If True, return per pixel level log prob.

        Returns
        -------
        dist
            A quantized logistic mixture `tfp.distribution` over the input data.
        """
        mixture_distribution = categorical.Categorical(logits=component_logits)

        # Convert distribution parameters for pixel values in
        # `[self._low, self._high]` for use with `QuantizedDistribution`
        locs = self._low + 0.5 * (self._high - self._low) * (locs + 1.0)
        scales *= 0.5 * (self._high - self._low)
        logistic_dist = quantized_distribution.QuantizedDistribution(
            distribution=transformed_distribution.TransformedDistribution(
                distribution=logistic.Logistic(loc=locs, scale=scales),
                bijector=Shift(shift=tf.cast(-0.5, self.dtype)),
            ),
            low=self._low,
            high=self._high,
        )

        # mixture with logistics for the loc and scale on each pixel for each component
        dist = mixture_same_family.MixtureSameFamily(
            mixture_distribution=mixture_distribution,
            components_distribution=independent.Independent(logistic_dist, reinterpreted_batch_ndims=1),
        )
        if return_per_feature:
            return dist
        else:
            return independent.Independent(dist, reinterpreted_batch_ndims=2)

    def _log_prob(self, value, conditional_input=None, training=None, return_per_feature=False):
        """Log probability function with optional conditional input.

        Calculates the log probability of a batch of data under the modeled
        distribution (or conditional distribution, if conditional input is
        provided).

        Parameters
        ----------
        value
            `Tensor` or Numpy array of image data. May have leading batch
            dimension(s), which must broadcast to the leading batch dimensions of
            `conditional_input`.
        conditional_input
            `Tensor` on which to condition the distribution (e.g.
            class labels), or `None`. May have leading batch dimension(s), which
            must broadcast to the leading batch dimensions of `value`.
        training
            `bool` or `None`. If `bool`, it controls the dropout layer,
            where `True` implies dropout is active. If `None`, it defaults to
            `keras.backend.learning_phase()`.
        return_per_feature
            `bool`. If True, return per pixel level log prob.

        Returns
        -------
        log_prob_values: `Tensor`.
        """
        # Determine the batch shape of the input images
        image_batch_shape = prefer_static.shape(value)[:-3]

        # Broadcast `value` and `conditional_input` to the same batch_shape
        if conditional_input is None:
            image_batch_and_conditional_shape = image_batch_shape
        else:
            conditional_input = tf.convert_to_tensor(conditional_input)
            conditional_input_shape = prefer_static.shape(conditional_input)
            conditional_batch_rank = prefer_static.rank(conditional_input) - tensorshape_util.rank(
                self.conditional_shape
            )
            conditional_batch_shape = conditional_input_shape[:conditional_batch_rank]

            image_batch_and_conditional_shape = prefer_static.broadcast_shape(
                image_batch_shape, conditional_batch_shape
            )
            conditional_input = tf.broadcast_to(
                conditional_input,
                prefer_static.concat([image_batch_and_conditional_shape, self.conditional_shape], axis=0),
            )
            value = tf.broadcast_to(
                value,
                prefer_static.concat([image_batch_and_conditional_shape, self.event_shape], axis=0),
            )

            # Flatten batch dimension for input to Keras model
            conditional_input = tf.reshape(
                conditional_input,
                prefer_static.concat([(-1,), self.conditional_shape], axis=0),
            )

        value = tf.reshape(value, prefer_static.concat([(-1,), self.event_shape], axis=0))

        transformed_value = (2.0 * (value - self._low) / (self._high - self._low)) - 1.0
        inputs = transformed_value if conditional_input is None else [transformed_value, conditional_input]

        params = self.network(inputs, training=training)

        num_channels = self.event_shape[-1]
        if num_channels == 1:
            component_logits, locs, scales = params
        else:
            # If there is more than one channel, we create a linear autoregressive
            # dependency among the location parameters of the channels of a single
            # pixel (the scale parameters within a pixel are independent). For a pixel
            # with R/G/B channels, the `r`, `g`, and `b` saturation values are
            # distributed as:
            #
            # r ~ Logistic(loc_r, scale_r)
            # g ~ Logistic(coef_rg * r + loc_g, scale_g)
            # b ~ Logistic(coef_rb * r + coef_gb * g + loc_b, scale_b)
            # on the coefficients instead of split/multiply/concat
            component_logits, locs, scales, coeffs = params
            num_coeffs = num_channels * (num_channels - 1) // 2
            loc_tensors = tf.split(locs, num_channels, axis=-1)
            coef_tensors = tf.split(coeffs, num_coeffs, axis=-1)
            channel_tensors = tf.split(value, num_channels, axis=-1)

            coef_count = 0
            for i in range(num_channels):
                channel_tensors[i] = channel_tensors[i][..., tf.newaxis, :]
                for j in range(i):
                    loc_tensors[i] += channel_tensors[j] * coef_tensors[coef_count]
                    coef_count += 1
            locs = tf.concat(loc_tensors, axis=-1)

        dist = self._make_mixture_dist(component_logits, locs, scales, return_per_feature=return_per_feature)
        log_px = dist.log_prob(value)
        if return_per_feature:
            return log_px
        else:
            return tf.reshape(log_px, image_batch_and_conditional_shape)

    def _sample_n(self, n, seed=None, conditional_input=None, training=False):
        """Samples from the distribution, with optional conditional input.

        Parameters
        ----------
        n
            `int`, number of samples desired.
        seed
            `int`, seed for RNG. Setting a random seed enforces reproducibility
            of the samples between sessions (not within a single session).
        conditional_input
            `Tensor` on which to condition the distribution (e.g.
            class labels), or `None`.
        training
            `bool` or `None`. If `bool`, it controls the dropout layer,
            where `True` implies dropout is active. If `None`, it defers to Keras'
            handling of train/eval status.

        Returns
        -------
        samples
            a `Tensor` of shape `[n, height, width, num_channels]`.
        """
        if conditional_input is not None:
            conditional_input = tf.convert_to_tensor(conditional_input, dtype=self.dtype)
            conditional_event_rank = tensorshape_util.rank(self.conditional_shape)
            conditional_input_shape = prefer_static.shape(conditional_input)
            conditional_sample_rank = prefer_static.rank(conditional_input) - conditional_event_rank

            # If `conditional_input` has no sample dimensions, prepend a sample
            # dimension
            if conditional_sample_rank == 0:
                conditional_input = conditional_input[tf.newaxis, ...]
                conditional_sample_rank = 1

            # Assert that the conditional event shape in the `PixelCnnNetwork` is the
            # same as that implied by `conditional_input`.
            conditional_event_shape = conditional_input_shape[conditional_sample_rank:]
            with tf.control_dependencies([tf.assert_equal(self.conditional_shape, conditional_event_shape)]):
                conditional_sample_shape = conditional_input_shape[:conditional_sample_rank]
                repeat = n // prefer_static.reduce_prod(conditional_sample_shape)
                h = tf.reshape(
                    conditional_input,
                    prefer_static.concat([(-1,), self.conditional_shape], axis=0),
                )
                h = tf.tile(
                    h,
                    prefer_static.pad(
                        [repeat],
                        paddings=[[0, conditional_event_rank]],
                        constant_values=1,
                    ),
                )

        samples_0 = tf.random.uniform(
            prefer_static.concat([(n,), self.event_shape], axis=0),
            minval=-1.0,
            maxval=1.0,
            dtype=self.dtype,
            seed=seed,
        )
        inputs = samples_0 if conditional_input is None else [samples_0, h]
        params_0 = self.network(inputs, training=training)
        samples_0 = self._sample_channels(*params_0, seed=seed)

        image_height, image_width, _ = tensorshape_util.as_list(self.event_shape)

        def loop_body(index, samples):
            """Loop for iterative pixel sampling.

            Parameters
            ----------
            index
                0D `Tensor` of type `int32`. Index of the current pixel.
            samples
              4D `Tensor`. Images with pixels sampled in raster order, up to
              pixel `[index]`, with dimensions `[batch_size, height, width,
              num_channels]`.

            Returns
            -------
            samples
                4D `Tensor`. Images with pixels sampled in raster order, up to \
                and including pixel `[index]`, with dimensions `[batch_size, height, \
                width, num_channels]`.
            """
            inputs = samples if conditional_input is None else [samples, h]
            params = self.network(inputs, training=training)
            samples_new = self._sample_channels(*params, seed=seed)

            # Update the current pixel
            samples = tf.transpose(samples, [1, 2, 3, 0])
            samples_new = tf.transpose(samples_new, [1, 2, 3, 0])
            row, col = index // image_width, index % image_width
            updates = samples_new[row, col, ...][tf.newaxis, ...]
            samples = tf.tensor_scatter_nd_update(samples, [[row, col]], updates)
            samples = tf.transpose(samples, [3, 0, 1, 2])

            return index + 1, samples

        index0 = tf.zeros([], dtype=tf.int32)

        # Construct the while loop for sampling
        total_pixels = image_height * image_width
        loop_cond = lambda ind, _: tf.less(ind, total_pixels)  # noqa: E731
        init_vars = (index0, samples_0)
        _, samples = tf.while_loop(loop_cond, loop_body, init_vars, parallel_iterations=1)

        transformed_samples = self._low + 0.5 * (self._high - self._low) * (samples + 1.0)
        return tf.round(transformed_samples)

    def _sample_channels(self, component_logits, locs, scales, coeffs=None, seed=None):
        """Sample a single pixel-iteration and apply channel conditioning.

        Parameters
        ----------
        component_logits
            4D `Tensor` of logits for the Categorical distribution
            over Quantized Logistic mixture components. Dimensions are `[batch_size,
            height, width, num_logistic_mix]`.
        locs
            4D `Tensor` of location parameters for the Quantized Logistic
            mixture components. Dimensions are `[batch_size, height, width,
            num_logistic_mix, num_channels]`.
        scales
            4D `Tensor` of location parameters for the Quantized Logistic
            mixture components. Dimensions are `[batch_size, height, width,
            num_logistic_mix, num_channels]`.
        coeffs
            4D `Tensor` of coefficients for the linear dependence among color
            channels, or `None` if there is only one channel. Dimensions are
            `[batch_size, height, width, num_logistic_mix, num_coeffs]`, where
            `num_coeffs = num_channels * (num_channels - 1) // 2`.
        seed
            `int`, random seed.

        Returns
        -------
        samples
            4D `Tensor` of sampled image data with autoregression among \
            channels. Dimensions are `[batch_size, height, width, num_channels]`.
        """
        num_channels = self.event_shape[-1]

        # sample mixture components once for the entire pixel
        component_dist = categorical.Categorical(logits=component_logits)
        mask = tf.one_hot(indices=component_dist.sample(seed=seed), depth=self._num_logistic_mix)
        mask = tf.cast(mask[..., tf.newaxis], self.dtype)

        # apply mixture component mask and separate out RGB parameters
        masked_locs = tf.reduce_sum(locs * mask, axis=-2)
        loc_tensors = tf.split(masked_locs, num_channels, axis=-1)
        masked_scales = tf.reduce_sum(scales * mask, axis=-2)
        scale_tensors = tf.split(masked_scales, num_channels, axis=-1)

        if coeffs is not None:
            num_coeffs = num_channels * (num_channels - 1) // 2
            masked_coeffs = tf.reduce_sum(coeffs * mask, axis=-2)
            coef_tensors = tf.split(masked_coeffs, num_coeffs, axis=-1)

        channel_samples = []
        coef_count = 0
        for i in range(num_channels):
            loc = loc_tensors[i]
            for c in channel_samples:
                loc += c * coef_tensors[coef_count]
                coef_count += 1

            logistic_samp = logistic.Logistic(loc=loc, scale=scale_tensors[i]).sample(seed=seed)
            logistic_samp = tf.clip_by_value(logistic_samp, -1.0, 1.0)
            channel_samples.append(logistic_samp)

        return tf.concat(channel_samples, axis=-1)

    def _batch_shape(self):
        return tf.TensorShape([])

    def _event_shape(self):
        return tf.TensorShape(self.image_shape)


class _PixelCNNNetwork(keras.layers.Layer):
    """Keras `Layer` to parameterize a Pixel CNN++ distribution.
    This is a Keras implementation of the Pixel CNN++ network, as described in
    Salimans et al. (2017)[1] and van den Oord et al. (2016)[2].
    (https://github.com/openai/pixel-cnn).
    #### References
    [1]: Tim Salimans, Andrej Karpathy, Xi Chen, and Diederik P. Kingma.
       PixelCNN++: Improving the PixelCNN with Discretized Logistic Mixture
       Likelihood and Other Modifications. In _International Conference on
       Learning Representations_, 2017.
       https://pdfs.semanticscholar.org/9e90/6792f67cbdda7b7777b69284a81044857656.pdf
       Additional details at https://github.com/openai/pixel-cnn
    [2]: Aaron van den Oord, Nal Kalchbrenner, Oriol Vinyals, Lasse Espeholt,
       Alex Graves, and Koray Kavukcuoglu. Conditional Image Generation with
       PixelCNN Decoders. In _30th Conference on Neural Information Processing
       Systems_, 2016.
       https://papers.nips.cc/paper/6527-conditional-image-generation-with-pixelcnn-decoders.pdf.
    """

    def __init__(
        self,
        dropout_p: float = 0.5,
        num_resnet: int = 5,
        num_hierarchies: int = 3,
        num_filters: int = 160,
        num_logistic_mix: int = 10,
        receptive_field_dims: tuple = (3, 3),
        resnet_activation: str = "concat_elu",
        l2_weight: float = 0.0,
        use_weight_norm: bool = True,
        use_data_init: bool = True,
        dtype=tf.float32,
    ) -> None:
        """Initialize the neural network for the Pixel CNN++ distribution.

        Parameters
        ----------
        dropout_p
            `float`, the dropout probability. Should be between 0 and 1.
        num_resnet
            `int`, the number of layers (shown in Figure 2 of [2]) within
            each highest-level block of Figure 2 of [1].
        num_hierarchies
            `int`, the number of hightest-level blocks (separated by
            expansions/contractions of dimensions in Figure 2 of [1].)
        num_filters
            `int`, the number of convolutional filters.
        num_logistic_mix
            `int`, number of components in the logistic mixture
            distribution.
        receptive_field_dims
            `tuple`, height and width in pixels of the receptive
            field of the convolutional layers above and to the left of a given
            pixel. The width (second element of the tuple) should be odd. Figure 1
            (middle) of [2] shows a receptive field of (3, 5) (the row containing
            the current pixel is included in the height). The default of (3, 3) was
            used to produce the results in [1].
        resnet_activation
            `string`, the type of activation to use in the resnet
            blocks. May be 'concat_elu', 'elu', or 'relu'.
        l2_weight
            `float`, the L2 regularization weight.
        use_weight_norm
            `bool`, if `True` then use weight normalization.
        use_data_init
            `bool`, if `True` then use data-dependent initialization
            (has no effect if `use_weight_norm` is `False`).
        dtype
            Data type of the layer.
        """
        super().__init__(dtype=dtype)
        self._dropout_p = dropout_p
        self._num_resnet = num_resnet
        self._num_hierarchies = num_hierarchies
        self._num_filters = num_filters
        self._num_logistic_mix = num_logistic_mix
        self._receptive_field_dims = receptive_field_dims  # first set desired receptive field, then infer kernel
        self._resnet_activation = resnet_activation
        self._l2_weight = l2_weight

        if use_weight_norm:

            def layer_wrapper(layer):
                def wrapped_layer(*args, **kwargs):
                    return WeightNorm(layer(*args, **kwargs), data_init=use_data_init)

                return wrapped_layer

            self._layer_wrapper = layer_wrapper
        else:
            self._layer_wrapper = lambda layer: layer

    def build(self, input_shape):
        dtype = self.dtype
        if len(input_shape) == 2:
            batch_image_shape, batch_conditional_shape = input_shape
            conditional_input = keras.layers.Input(shape=batch_conditional_shape[1:], dtype=dtype)
        else:
            batch_image_shape = input_shape
            conditional_input = None

        image_shape = batch_image_shape[1:]
        image_input = keras.layers.Input(shape=image_shape, dtype=dtype)

        if self._resnet_activation == "concat_elu":
            activation = keras.layers.Lambda(lambda x: tf.nn.elu(tf.concat([x, -x], axis=-1)), dtype=dtype)
        else:
            activation = keras.activations.get(self._resnet_activation)

        # Define layers with default inputs and layer wrapper applied
        Conv2D = functools.partial(  # pylint:disable=invalid-name
            self._layer_wrapper(keras.layers.Convolution2D),
            filters=self._num_filters,
            padding="same",
            kernel_regularizer=keras.regularizers.l2(self._l2_weight),
            dtype=dtype,
        )

        Dense = functools.partial(  # pylint:disable=invalid-name
            self._layer_wrapper(keras.layers.Dense),
            kernel_regularizer=keras.regularizers.l2(self._l2_weight),
            dtype=dtype,
        )

        Conv2DTranspose = functools.partial(  # pylint:disable=invalid-name
            self._layer_wrapper(keras.layers.Conv2DTranspose),
            filters=self._num_filters,
            padding="same",
            strides=(2, 2),
            kernel_regularizer=keras.regularizers.l2(self._l2_weight),
            dtype=dtype,
        )

        rows, cols = self._receptive_field_dims

        # Define the dimensions of the valid (unmasked) areas of the layer kernels
        # for stride 1 convolutions in the internal layers.
        kernel_valid_dims = {
            "vertical": (rows - 1, cols),  # vertical stack
            "horizontal": (2, cols // 2 + 1),
        }  # horizontal stack

        # Define the size of the kernel necessary to center the current pixel
        # correctly for stride 1 convolutions in the internal layers.
        kernel_sizes = {"vertical": (2 * rows - 3, cols), "horizontal": (3, cols)}

        # Make the kernel constraint functions for stride 1 convolutions in internal
        # layers.
        kernel_constraints = {
            k: _make_kernel_constraint(kernel_sizes[k], (0, v[0]), (0, v[1])) for k, v in kernel_valid_dims.items()
        }

        # Build the initial vertical stack/horizontal stack convolutional layers,
        # as shown in Figure 1 of [2]. The receptive field of the initial vertical
        # stack layer is a rectangular area centered above the current pixel.
        vertical_stack_init = Conv2D(
            kernel_size=(2 * rows - 1, cols),
            kernel_constraint=_make_kernel_constraint((2 * rows - 1, cols), (0, rows - 1), (0, cols)),
        )(image_input)

        # In Figure 1 [2], the receptive field of the horizontal stack is
        # illustrated as the pixels in the same row and to the left of the current
        # pixel. [1] increases the height of this receptive field from one pixel to
        # two (`horizontal_stack_left`) and additionally includes a subset of the
        # row of pixels centered above the current pixel (`horizontal_stack_up`).
        horizontal_stack_up = Conv2D(
            kernel_size=(3, cols),
            kernel_constraint=_make_kernel_constraint((3, cols), (0, 1), (0, cols)),
        )(image_input)

        horizontal_stack_left = Conv2D(
            kernel_size=(3, cols),
            kernel_constraint=_make_kernel_constraint((3, cols), (0, 2), (0, cols // 2)),
        )(image_input)

        horizontal_stack_init = keras.layers.add([horizontal_stack_up, horizontal_stack_left], dtype=dtype)

        layer_stacks = {
            "vertical": [vertical_stack_init],
            "horizontal": [horizontal_stack_init],
        }

        # Build the downward pass of the U-net (left-hand half of Figure 2 of [1]).
        # Each `i` iteration builds one of the highest-level blocks (identified as
        # 'Sequence of 6 layers' in the figure, consisting of `num_resnet=5` stride-
        # 1 layers, and one stride-2 layer that contracts the height/width
        # dimensions). The `_` iterations build the stride 1 layers. The layers of
        # the downward pass are stored in lists, since we'll later need them to make
        # skip-connections to layers in the upward pass of the U-net (the skip-
        # connections are represented by curved lines in Figure 2 [1]).
        for i in range(self._num_hierarchies):
            for _ in range(self._num_resnet):
                # Build a layer shown in Figure 2 of [2]. The 'vertical' iteration
                # builds the layers in the left half of the figure, and the 'horizontal'
                # iteration builds the layers in the right half.
                for stack in ["vertical", "horizontal"]:
                    input_x = layer_stacks[stack][-1]
                    x = activation(input_x)
                    x = Conv2D(
                        kernel_size=kernel_sizes[stack],
                        kernel_constraint=kernel_constraints[stack],
                    )(x)

                    # Add the vertical-stack layer to the horizontal-stack layer
                    if stack == "horizontal":
                        h = activation(layer_stacks["vertical"][-1])
                        h = Dense(self._num_filters)(h)
                        x = keras.layers.add([h, x], dtype=dtype)

                    x = activation(x)
                    x = keras.layers.Dropout(self._dropout_p, dtype=dtype)(x)
                    x = Conv2D(
                        filters=2 * self._num_filters,
                        kernel_size=kernel_sizes[stack],
                        kernel_constraint=kernel_constraints[stack],
                    )(x)

                    if conditional_input is not None:
                        h_projection = _build_and_apply_h_projection(conditional_input, self._num_filters, dtype=dtype)
                        x = keras.layers.add([x, h_projection], dtype=dtype)

                    x = _apply_sigmoid_gating(x)

                    # Add a residual connection from the layer's input.
                    out = keras.layers.add([input_x, x], dtype=dtype)
                    layer_stacks[stack].append(out)

            if i < self._num_hierarchies - 1:
                # Build convolutional layers that contract the height/width dimensions
                # on the downward pass between each set of layers (e.g. contracting from
                # 32x32 to 16x16 in Figure 2 of [1]).
                for stack in ["vertical", "horizontal"]:
                    # Define kernel dimensions/masking to maintain the autoregressive property.
                    x = layer_stacks[stack][-1]
                    h, w = kernel_valid_dims[stack]
                    kernel_height = 2 * h
                    kernel_width = w + 1 if stack == "vertical" else 2 * w
                    kernel_size = (kernel_height, kernel_width)
                    kernel_constraint = _make_kernel_constraint(kernel_size, (0, h), (0, w))
                    x = Conv2D(
                        strides=(2, 2),
                        kernel_size=kernel_size,
                        kernel_constraint=kernel_constraint,
                    )(x)
                    layer_stacks[stack].append(x)

        # Upward pass of the U-net (right-hand half of Figure 2 of [1]). We stored
        # the layers of the downward pass in a list, in order to access them to make
        # skip-connections to the upward pass. For the upward pass, we need to keep
        # track of only the current layer, so we maintain a reference to the
        # current layer of the horizontal/vertical stack in the `upward_pass` dict.
        # The upward pass begins with the last layer of the downward pass.
        upward_pass = {key: stack.pop() for key, stack in layer_stacks.items()}

        # As with the downward pass, each `i` iteration builds a highest level block
        # in Figure 2 [1], and the `_` iterations build individual layers within the
        # block.
        for i in range(self._num_hierarchies):
            num_resnet = self._num_resnet if i == 0 else self._num_resnet + 1

            for _ in range(num_resnet):
                # Build a layer as shown in Figure 2 of [2], with a skip-connection
                # from the symmetric layer in the downward pass.
                for stack in ["vertical", "horizontal"]:
                    input_x = upward_pass[stack]
                    x_symmetric = layer_stacks[stack].pop()

                    x = activation(input_x)
                    x = Conv2D(
                        kernel_size=kernel_sizes[stack],
                        kernel_constraint=kernel_constraints[stack],
                    )(x)

                    # Include the vertical-stack layer of the upward pass in the layers
                    # to be added to the horizontal layer.
                    if stack == "horizontal":
                        x_symmetric = keras.layers.Concatenate(axis=-1, dtype=dtype)(
                            [upward_pass["vertical"], x_symmetric]
                        )

                    # Add a skip-connection from the symmetric layer in the downward
                    # pass to the layer `x` in the upward pass.
                    h = activation(x_symmetric)
                    h = Dense(self._num_filters)(h)
                    x = keras.layers.add([h, x], dtype=dtype)

                    x = activation(x)
                    x = keras.layers.Dropout(self._dropout_p, dtype=dtype)(x)
                    x = Conv2D(
                        filters=2 * self._num_filters,
                        kernel_size=kernel_sizes[stack],
                        kernel_constraint=kernel_constraints[stack],
                    )(x)

                    if conditional_input is not None:
                        h_projection = _build_and_apply_h_projection(conditional_input, self._num_filters, dtype=dtype)
                        x = keras.layers.add([x, h_projection], dtype=dtype)

                    x = _apply_sigmoid_gating(x)
                    upward_pass[stack] = keras.layers.add([input_x, x], dtype=dtype)

            # Define deconvolutional layers that expand height/width dimensions on the
            # upward pass (e.g. expanding from 8x8 to 16x16 in Figure 2 of [1]), with
            # the correct kernel dimensions/masking to maintain the autoregressive
            # property.
            if i < self._num_hierarchies - 1:
                for stack in ["vertical", "horizontal"]:
                    h, w = kernel_valid_dims[stack]
                    kernel_height = 2 * h - 2
                    if stack == "vertical":
                        kernel_width = w + 1
                        kernel_constraint = _make_kernel_constraint(
                            (kernel_height, kernel_width),
                            (h - 2, kernel_height),
                            (0, w),
                        )
                    else:
                        kernel_width = 2 * w - 2
                        kernel_constraint = _make_kernel_constraint(
                            (kernel_height, kernel_width),
                            (h - 2, kernel_height),
                            (w - 2, kernel_width),
                        )

                    x = upward_pass[stack]
                    x = Conv2DTranspose(
                        kernel_size=(kernel_height, kernel_width),
                        kernel_constraint=kernel_constraint,
                    )(x)
                    upward_pass[stack] = x

        x_out = keras.layers.ELU(dtype=dtype)(upward_pass["horizontal"])

        # Build final Dense/Reshape layers to output the correct number of
        # parameters per pixel.
        num_channels = tensorshape_util.as_list(image_shape)[-1]
        num_coeffs = num_channels * (num_channels - 1) // 2  # alpha, beta, gamma in eq.3 of paper
        num_out = num_channels * 2 + num_coeffs + 1  # mu, s + alpha, beta, gamma + 1 (mixture weight)
        num_out_total = num_out * self._num_logistic_mix
        params = Dense(num_out_total)(x_out)
        params = tf.reshape(
            params,
            prefer_static.concat(  # [-1,H,W,nb mixtures, params per mixture]
                [[-1], image_shape[:-1], [self._num_logistic_mix, num_out]], axis=0
            ),
        )

        # If there is one color channel, split the parameters into a list of three
        # output `Tensor`s: (1) component logits for the Quantized Logistic mixture
        # distribution, (2) location parameters for each component, and (3) scale
        # parameters for each component. If there is more than one color channel,
        # return a fourth `Tensor` for the coefficients for the linear dependence
        # among color channels (e.g. alpha, beta, gamma).
        # [logits, mu, s, linear dependence]
        splits = 3 if num_channels == 1 else [1, num_channels, num_channels, num_coeffs]
        outputs = tf.split(params, splits, axis=-1)

        # Squeeze singleton dimension from component logits
        outputs[0] = tf.squeeze(outputs[0], axis=-1)

        # Ensure scales are positive and do not collapse to near-zero
        outputs[2] = tf.nn.softplus(outputs[2]) + tf.cast(tf.exp(-7.0), self.dtype)

        inputs = image_input if conditional_input is None else [image_input, conditional_input]
        self._network = keras.Model(inputs=inputs, outputs=outputs)
        super().build(input_shape)

    def call(self, inputs, training=None):
        """Call the Pixel CNN network model.

        Parameters
        ----------
        inputs
            4D `Tensor` of image data with dimensions [batch size, height,
            width, channels] or a 2-element `list`. If `list`, the first element is
            the 4D image `Tensor` and the second element is a `Tensor` with
            conditional input data (e.g. VAE encodings or class labels) with the
            same leading batch dimension as the image `Tensor`.
        training
            `bool` or `None`. If `bool`, it controls the dropout layer,
            where `True` implies dropout is active. If `None`, it it defaults to
            `keras.backend.learning_phase()`

        Returns
        -------
        outputs
            a 3- or 4-element `list` of `Tensor`s in the following order: \
            component_logits: 4D `Tensor` of logits for the Categorical distribution \
            over Quantized Logistic mixture components. Dimensions are \
            `[batch_size, height, width, num_logistic_mix]`.
        locs
            4D `Tensor` of location parameters for the Quantized Logistic \
            mixture components. Dimensions are `[batch_size, height, width, \
            num_logistic_mix, num_channels]`.
        scales
            4D `Tensor` of location parameters for the Quantized Logistic \
            mixture components. Dimensions are `[batch_size, height, width, \
            num_logistic_mix, num_channels]`.
        coeffs
            4D `Tensor` of coefficients for the linear dependence among \
            color channels, included only if the image has more than one channel. \
            Dimensions are `[batch_size, height, width, num_logistic_mix, \
            num_coeffs]`, where `num_coeffs = num_channels * (num_channels - 1) // 2`.
        """
        return self._network(inputs, training=training)


def _make_kernel_constraint(kernel_size, valid_rows, valid_columns):
    """Make the masking function for layer kernels."""
    mask = np.zeros(kernel_size)
    lower, upper = valid_rows
    left, right = valid_columns
    mask[lower:upper, left:right] = 1.0
    mask = mask[:, :, np.newaxis, np.newaxis]
    return lambda x: x * mask


def _build_and_apply_h_projection(h, num_filters, dtype):
    """Project the conditional input."""
    h = keras.layers.Flatten(dtype=dtype)(h)
    h_projection = keras.layers.Dense(2 * num_filters, kernel_initializer="random_normal", dtype=dtype)(h)
    return h_projection[..., tf.newaxis, tf.newaxis, :]


def _apply_sigmoid_gating(x):
    """Apply the sigmoid gating in Figure 2 of [2]."""
    activation_tensor, gate_tensor = tf.split(x, 2, axis=-1)
    sigmoid_gate = tf.sigmoid(gate_tensor)
    return keras.layers.multiply([sigmoid_gate, activation_tensor], dtype=x.dtype)



from importlib.util import find_spec

from daml._internal.detectors.drift.base import LastSeenUpdate, ReservoirSamplingUpdate
from daml._internal.detectors.drift.cvm import DriftCVM
from daml._internal.detectors.drift.ks import DriftKS
from daml._internal.detectors.duplicates import Duplicates
from daml._internal.detectors.linter import Linter

__all__ = ["Duplicates", "Linter", "DriftCVM", "DriftKS", "LastSeenUpdate", "ReservoirSamplingUpdate"]

if find_spec("torch") is not None:  # pragma: no cover
    from daml._internal.detectors.drift.mmd import DriftMMD
    from daml._internal.detectors.drift.torch import GaussianRBF, preprocess_drift
    from daml._internal.detectors.drift.uncertainty import DriftUncertainty

    __all__ += ["DriftMMD", "GaussianRBF", "DriftUncertainty", "preprocess_drift"]

if find_spec("tensorflow") is not None and find_spec("tensorflow_probability") is not None:  # pragma: no cover
    from daml._internal.detectors.ood.ae import OOD_AE
    from daml._internal.detectors.ood.aegmm import OOD_AEGMM
    from daml._internal.detectors.ood.base import OODScore
    from daml._internal.detectors.ood.llr import OOD_LLR
    from daml._internal.detectors.ood.vae import OOD_VAE
    from daml._internal.detectors.ood.vaegmm import OOD_VAEGMM

    __all__ += ["OOD_AE", "OOD_AEGMM", "OOD_LLR", "OODScore", "OOD_VAE", "OOD_VAEGMM"]

del find_spec



from typing import Dict, Iterable, List, NamedTuple, Tuple, Union, cast

import numpy as np
from scipy.cluster.hierarchy import linkage
from scipy.spatial.distance import pdist, squareform


def extend_linkage(link_arr: np.ndarray) -> np.ndarray:
    """
    Adds a column to the linkage matrix link_arr that tracks the new id assigned
    to each row

    Parameters
    ----------
    link_arr : np.ndarray
        linkage matrix

    Returns
    -------
    np.ndarray
        linkage matrix with adjusted shape, new shape (link_arr.shape[0], link_arr.shape[1]+1)
    """
    # Adjusting linkage matrix to accommodate renumbering
    rows, cols = link_arr.shape
    arr = np.zeros((rows, cols + 1))
    arr[:, :-1] = link_arr
    arr[:, -1] = np.arange(rows + 1, 2 * rows + 1)

    return arr


class Cluster:
    __slots__ = "merged", "samples", "sample_dist", "is_copy", "count", "dist_avg", "dist_std", "out1", "out2"

    def __init__(self, merged: int, samples: np.ndarray, sample_dist: Union[float, np.ndarray], is_copy: bool = False):
        self.merged = merged
        self.samples = np.array(samples, dtype=np.int32)
        self.sample_dist = np.array([sample_dist] if np.isscalar(sample_dist) else sample_dist)
        self.is_copy = is_copy

        dist = float(self.sample_dist[-1])

        self.count = len(self.samples)
        if is_copy:
            self.dist_avg = 0.0
            self.dist_std = 0.0
            self.out1 = False
            self.out2 = False
        else:
            self.dist_avg = float(np.mean(self.sample_dist))
            self.dist_std = float(np.std(self.sample_dist)) if len(self.sample_dist) > 1 else 1e-5
            out1 = self.dist_avg + self.dist_std
            out2 = out1 + self.dist_std
            self.out1 = dist > out1
            self.out2 = dist > out2

    def copy(self) -> "Cluster":
        return Cluster(False, self.samples, self.sample_dist, True)

    def __repr__(self) -> str:
        _params = {
            "merged": self.merged,
            "samples": self.samples,
            "sample_dist": self.sample_dist,
            "is_copy": self.is_copy,
        }
        return f"{self.__class__.__name__}(**{repr(_params)})"


class ClusterPosition(NamedTuple):
    """Keeps track of a cluster's level and ID"""

    level: int
    cid: int


class ClusterMergeEntry:
    __slots__ = "level", "outer_cluster", "inner_cluster", "status"

    def __init__(self, level: int, outer_cluster: int, inner_cluster: int, status: int):
        self.level = level
        self.outer_cluster = outer_cluster
        self.inner_cluster = inner_cluster
        self.status = status

    def __lt__(self, value: "ClusterMergeEntry") -> bool:
        return self.level.__lt__(value.level)

    def __gt__(self, value: "ClusterMergeEntry") -> bool:
        return self.level.__gt__(value.level)


class Clusterer:
    """
    Uses hierarchical clustering to flag dataset properties of interest like outliers and duplicates

    Parameters
    ----------
    dataset : np.ndarray
        An array of images or image embeddings to perform clustering
    """

    def __init__(self, dataset: np.ndarray):
        # Allows an update to dataset to reset the state rather than instantiate a new class
        self._on_init(dataset)

    def _on_init(self, dataset: np.ndarray):
        self._validate_data(dataset)
        self._data: np.ndarray = dataset
        self._num_samples = len(dataset)

        self._darr: np.ndarray = pdist(dataset, metric="euclidean")
        self._sqdmat: np.ndarray = squareform(self._darr)
        self._larr: np.ndarray = extend_linkage(linkage(self._darr))
        self._max_clusters: int = np.count_nonzero(self._larr[:, 3] == 2)
        self._max_level: int = 1

        min_num = int(self._num_samples * 0.05)
        self._min_num_samples_per_cluster = min(max(2, min_num), 100)

        self._clusters: Dict[int, Dict[int, Cluster]] = {}  # Dictionary to store clusters
        self._last_good_merge_levels: Dict = {}

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, x: np.ndarray):
        self._on_init(x)

    @property
    def clusters(self) -> Dict[int, Dict[int, Cluster]]:
        if not self._clusters:
            self._clusters = self._create_clusters()
        return self._clusters

    @property
    def last_good_merge_levels(self) -> Dict[int, int]:
        if not self._last_good_merge_levels:
            self._last_good_merge_levels = self._get_last_merge_levels()
        return self._last_good_merge_levels

    @classmethod
    def _validate_data(cls, x: np.ndarray):
        """Checks that the data has the correct size, shape, and format"""
        if not isinstance(x, np.ndarray):
            raise TypeError(f"Data should be of type np.ndarray; got {type(x)}")

        if x.ndim != 2:
            raise ValueError(
                f"Data should only have 2 dimensions; got {x.ndim}. Data should be flattened before being input"
            )
        samples, features = x.shape  # Due to above check, we know shape has a length of 2
        if samples < 2:
            raise ValueError(f"Data should have at least 2 samples; got {samples}")
        if features < 1:
            raise ValueError(f"Samples should have at least 1 feature; got {features}")

    def _create_clusters(self) -> Dict[int, Dict[int, Cluster]]:
        """Generates clusters based on linkage matrix"""
        next_cluster_id = 0
        cluster_map: Dict[int, ClusterPosition] = {}  # Dictionary to associate new cluster ids with actual clusters
        clusters: Dict[int, Dict[int, Cluster]] = {}

        # Walking through the linkage array to generate clusters
        for arr_i in self._larr:
            left_id = int(arr_i[0])
            right_id = int(arr_i[1])
            sample_dist = np.array([arr_i[2]], dtype=np.float16)
            merged = False

            # Determine if the id is already associated with a cluster
            left = cluster_map.get(left_id)
            right = cluster_map.get(right_id)

            if left and right:
                merged = max([left.cid, right.cid])
                lc = clusters[left.level][left.cid]
                rc = clusters[right.level][right.cid]
                left_first = len(lc.samples) >= len(rc.samples)
                samples = np.concatenate([lc.samples, rc.samples] if left_first else [rc.samples, lc.samples])
                sample_dist = np.concatenate([rc.sample_dist, lc.sample_dist, sample_dist])
                level, cid = max(left.level, right.level) + 1, min(left.cid, right.cid)

                # Only tracking the levels in which clusters merge for the cluster distance matrix
                self._max_level = max(self._max_level, left.level, right.level)
                # Update clusters to include previously skipped levels
                clusters = self._fill_levels(clusters, left, right)
            elif left or right:
                child, other_id = cast(Tuple[ClusterPosition, int], (left, right_id) if left else (right, left_id))
                cc = clusters[child.level][child.cid]
                samples = np.concatenate([cc.samples, [other_id]])
                sample_dist = np.concatenate([cc.sample_dist, sample_dist])
                level, cid = child.level + 1, child.cid
            else:
                samples = np.array([left_id, right_id], dtype=np.int32)
                level, cid = 0, next_cluster_id
                next_cluster_id += 1

            # Set the cluster and associate the linkage id with the cluster
            if level not in clusters:
                clusters[level] = {}

            clusters[level][cid] = Cluster(merged, samples, sample_dist)
            cluster_map[int(arr_i[-1])] = ClusterPosition(level, cid)

        return clusters

    def _fill_levels(
        self, clusters: Dict[int, Dict[int, Cluster]], left: ClusterPosition, right: ClusterPosition
    ) -> Dict[int, Dict[int, Cluster]]:
        # Sets each level's cluster info if it does not exist
        if left.level != right.level:
            (level, cid), max_level = (left, right[0]) if left[0] < right[0] else (right, left[0])
            cluster = clusters[level][cid].copy()
            for level_id in range(max_level, level, -1):
                clusters[level_id].setdefault(cid, cluster)
        return clusters

    def _get_cluster_distances(self) -> np.ndarray:
        """Calculates the minimum distances between clusters are each level"""
        # Must call self.clusters before creating matrix because
        # self._max_level and self._max_clusters are calculated by self.clusters property
        clusters = self.clusters
        # Cluster distance matrix
        cluster_matrix = np.full((self._max_level, self._max_clusters, self._max_clusters), -1.0, dtype=np.float32)

        for level, cluster_set in clusters.items():
            if level < self._max_level:
                cluster_ids = sorted(cluster_set.keys())
                for i, cluster_id in enumerate(cluster_ids):
                    cluster_matrix[level, cluster_id, cluster_id] = clusters[level][cluster_id].dist_avg
                    for int_id in range(i + 1, len(cluster_ids)):
                        compare_id = cluster_ids[int_id]
                        sample_a = clusters[level][cluster_id].samples
                        sample_b = clusters[level][compare_id].samples
                        min_mat = self._sqdmat[np.ix_(sample_a, sample_b)].min()
                        cluster_matrix[level, cluster_id, compare_id] = min_mat
                        cluster_matrix[level, compare_id, cluster_id] = min_mat

        return cluster_matrix

    def _calc_merge_indices(self, merge_mean: List[np.ndarray], intra_max: List[float]) -> np.ndarray:
        """
        Determine what clusters should be merged and return their indices
        """
        intra_max_uniques = np.unique(intra_max)
        intra_log_values = np.log(intra_max_uniques)
        two_std_all = intra_log_values.mean() + 2 * intra_log_values.std()
        merge_value = np.log(merge_mean)
        # Mask of indices we know we want to merge
        desired_merge = merge_value < two_std_all

        # List[Values] for indices we might want to merge
        check = merge_value[~desired_merge]
        # Check distance from value to 2 stds of all values
        check = np.abs((check - two_std_all) / two_std_all)
        # Mask List[Values < 1]
        mask = check < 1
        one_std_check = check[mask].mean() + check[mask].std()
        # Mask of indices that should also be merged
        mask2_vals = np.abs((merge_value - two_std_all) / two_std_all)
        mask2 = mask2_vals < one_std_check
        return np.logical_or(desired_merge, mask2)

    def _generate_merge_list(self, cluster_matrix: np.ndarray) -> List[ClusterMergeEntry]:
        """
        Runs through the clusters dictionary determining when clusters merge,
        and how close are those clusters when they merge.

        Parameters
        ----------
        cluster_matrix:
            The distance matrix for all clusters to all others

        Returns
        -------
        List[ClusterMergeEntry]:
            A list with each cluster's merge history
        """
        intra_max = []
        merge_mean = []
        merge_list: List[ClusterMergeEntry] = []

        for level, cluster_set in self.clusters.items():
            for outer_cluster, cluster in cluster_set.items():
                inner_cluster = cluster.merged
                if not inner_cluster:
                    continue
                # Extract necessary information
                num_samples = len(cluster.samples)
                out1 = cluster.out1
                out2 = cluster.out2

                # If outside 2-std or 1-std and larger than a minimum sized cluster, take the mean distance, else max
                aggregate_func = (
                    np.mean if out2 or (out1 and num_samples >= self._min_num_samples_per_cluster) else np.max
                )

                distances = cluster_matrix[:level, outer_cluster, inner_cluster]
                intra_distance = cluster_matrix[:, outer_cluster, outer_cluster]
                positive_mask = intra_distance >= 0
                intra_filtered = intra_distance[positive_mask]

                # TODO: Append now, take max over axis later?
                intra_max.append(np.max(intra_filtered))
                # Calculate the corresponding distance stats
                distance_stats_arr = aggregate_func(distances)
                merge_mean.append(distance_stats_arr)
                merge_list.append(ClusterMergeEntry(level, outer_cluster, inner_cluster, 0))

        all_merge_indices = self._calc_merge_indices(merge_mean=merge_mean, intra_max=intra_max)

        for i, is_mergeable in enumerate(all_merge_indices):
            merge_list[i].status = is_mergeable

        merge_list = sorted(merge_list, reverse=True)

        return merge_list

    def _get_last_merge_levels(self) -> Dict[int, int]:
        """
        Creates a dictionary for important cluster ids mapped to their last good merge level

        Returns
        -------
        Dict[int, int]
            A mapping of a cluster id to its last good merge level
        """
        last_merge_levels: Dict[int, int] = {}

        if self._max_clusters <= 1:
            last_merge_levels = {0: int(self._num_samples * 0.1)}
        else:
            cluster_matrix = self._get_cluster_distances()
            merge_list = self._generate_merge_list(cluster_matrix)
            for entry in merge_list:
                if not entry.status:
                    if entry.outer_cluster not in last_merge_levels:
                        last_merge_levels[entry.outer_cluster] = 0
                    if entry.inner_cluster not in last_merge_levels:
                        last_merge_levels[entry.inner_cluster] = 0
                    if last_merge_levels[entry.outer_cluster] > entry.level:
                        last_merge_levels[entry.outer_cluster] = entry.level - 1
                else:
                    if entry.outer_cluster in last_merge_levels:
                        last_merge_levels[entry.outer_cluster] = max(
                            last_merge_levels[entry.outer_cluster], entry.level
                        )

        return last_merge_levels

    def find_outliers(self, last_merge_levels: Dict[int, int]) -> Tuple[List[int], List[int]]:
        """
        Retrieves outliers based on when the sample was added to the cluster
        and how far it was from the cluster when it was added

        Parameters
        ----------
        last_merge_levels : Dict[int, int]
            A mapping of a cluster id to its last good merge level

        Returns
        -------
        Tuple[List[int], List[int]]
            The outliers and possible outliers as sorted lists of indices
        """
        outliers = set()
        possible_outliers = set()

        for level, cluster_set in self.clusters.items():
            for cluster_id, cluster in cluster_set.items():
                if not cluster.merged and cluster_id in last_merge_levels and level > last_merge_levels[cluster_id]:
                    if cluster.out2:
                        outliers.add(cluster.samples[-1])
                    elif cluster.out1 and len(cluster.samples) >= self._min_num_samples_per_cluster:
                        possible_outliers.add(cluster.samples[-1])
                    elif len(cluster.samples) < self._min_num_samples_per_cluster:
                        outliers.update(cluster.samples.tolist())

        return sorted(outliers), sorted(possible_outliers)

    def _sorted_union_find(self, index_groups: Iterable[Iterable[int]]) -> List[List[int]]:
        """Merges and sorts groups of indices that share any common index"""
        groups: List[List[int]] = []
        for indices in zip(*index_groups):
            indices = set(indices)
            temp = []
            for group in groups:
                if not set(group).isdisjoint(indices):
                    indices.update(group)
                else:
                    temp.append(group)
            temp.append(sorted(indices))
            groups = temp
        return sorted(groups)

    def find_duplicates(self, last_merge_levels: Dict[int, int]) -> Tuple[List[List[int]], List[List[int]]]:
        """
        Finds duplicate and near duplicate data based on the last good merge levels when building the cluster

        Parameters
        ----------
        last_merge_levels : Dict[int, int]
            A mapping of a cluster id to its last good merge level

        Returns
        -------
        Tuple[List[List[int]], List[List[int]]]
            The exact duplicates and near duplicates as lists of related indices
        """

        duplicates_std = []
        for cluster_id, level in last_merge_levels.items():
            samples = self.clusters[level][cluster_id].samples
            if len(samples) >= self._min_num_samples_per_cluster:
                duplicates_std.append(self.clusters[level][cluster_id].dist_std)
        diag_mask = np.ones_like(self._sqdmat, dtype=bool)
        np.fill_diagonal(diag_mask, 0)
        diag_mask = np.triu(diag_mask)

        exact_mask = self._sqdmat <= (np.mean(duplicates_std) / 100)
        exact_indices = np.nonzero(exact_mask & diag_mask)
        exact_dupes = self._sorted_union_find(exact_indices)

        near_mask = self._sqdmat <= np.mean(duplicates_std)
        near_indices = np.nonzero(near_mask & diag_mask & ~exact_mask)
        near_dupes = self._sorted_union_find(near_indices)

        return exact_dupes, near_dupes

    def evaluate(self):
        """Finds and flags indices of the data for outliers and duplicates

        Returns
        -------

        Dict[str, Union[List[int]], List[List[int]]]
            Dictionary containing list of outliers, potential outliers, duplicates, and near duplicates in keys
            "outliers", "potential_outliers", "duplicates", "near_duplicates" respectively
        """

        outliers, potential_outliers = self.find_outliers(self.last_good_merge_levels)
        duplicates, near_duplicates = self.find_duplicates(self.last_good_merge_levels)

        ret = {
            "outliers": outliers,
            "potential_outliers": potential_outliers,
            "duplicates": duplicates,
            "near_duplicates": near_duplicates,
        }

        return ret



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable, Literal, Optional, Tuple

import numpy as np
from scipy.stats import cramervonmises_2samp

from .base import BaseUnivariateDrift, UpdateStrategy, preprocess_x


class DriftCVM(BaseUnivariateDrift):
    """
    Cramér-von Mises (CVM) data drift detector, which tests for any change in the
    distribution of continuous univariate data. For multivariate data, a separate
    CVM test is applied to each feature, and the obtained p-values are aggregated
    via the Bonferroni or False Discovery Rate (FDR) corrections.

    Parameters
    ----------
    x_ref : np.ndarray
        Data used as reference distribution.
    p_val : float, default 0.05
        p-value used for significance of the statistical test for each feature.
        If the FDR correction method is used, this corresponds to the acceptable
        q-value.
    x_ref_preprocessed : bool, default False
        Whether the given reference data `x_ref` has been preprocessed yet. If
        `x_ref_preprocessed=True`, only the test data `x` will be preprocessed at
        prediction time. If `x_ref_preprocessed=False`, the reference data will also
        be preprocessed.
    update_x_ref : Optional[UpdateStrategy], default None
        Reference data can optionally be updated using an UpdateStrategy class. Update
        using the last n instances seen by the detector with
        :py:class:`daml.detectors.LastSeenUpdateStrategy`
        or via reservoir sampling with
        :py:class:`daml.detectors.ReservoirSamplingUpdateStrategy`.
    preprocess_fn : Optional[Callable[[np.ndarray], np.ndarray]], default None
        Function to preprocess the data before computing the data drift metrics.
        Typically a dimensionality reduction technique.
    correction : Literal["bonferroni", "fdr"], default "bonferroni"
        Correction type for multivariate data. Either 'bonferroni' or 'fdr' (False
        Discovery Rate).
    n_features
        Number of features used in the statistical test. No need to pass it if no
        preprocessing takes place. In case of a preprocessing step, this can also
        be inferred automatically but could be more expensive to compute.
    """

    def __init__(
        self,
        x_ref: np.ndarray,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preprocess_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        correction: Literal["bonferroni", "fdr"] = "bonferroni",
        n_features: Optional[int] = None,
    ) -> None:
        super().__init__(
            x_ref=x_ref,
            p_val=p_val,
            x_ref_preprocessed=x_ref_preprocessed,
            update_x_ref=update_x_ref,
            preprocess_fn=preprocess_fn,
            correction=correction,
            n_features=n_features,
        )

    @preprocess_x
    def score(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Performs the two-sample Cramér-von Mises test(s), computing the p-value and
        test statistic per feature.

        Parameters
        ----------
        x
            Batch of instances.

        Returns
        -------
        Feature level p-values and CVM statistics.
        """
        x = x.reshape(x.shape[0], -1)
        x_ref = self.x_ref.reshape(self.x_ref.shape[0], -1)
        p_val = np.zeros(self.n_features, dtype=np.float32)
        dist = np.zeros_like(p_val)
        for f in range(self.n_features):
            result = cramervonmises_2samp(x_ref[:, f], x[:, f], method="auto")
            p_val[f], dist[f] = result.pvalue, result.statistic
        return p_val, dist



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

import math
from typing import Callable, Optional, Tuple, Type, Union, cast

import keras.api._v2.keras as keras
import numpy as np
import tensorflow as tf
from keras.api._v2.keras import Sequential
from keras.api._v2.keras.layers import (
    Conv2D,
    Conv2DTranspose,
    Dense,
    Flatten,
    InputLayer,
    Reshape,
)
from tensorflow._api.v2.nn import relu, softmax, tanh

from daml._internal.models.tensorflow.autoencoder import AE, AEGMM, VAE, VAEGMM
from daml._internal.models.tensorflow.pixelcnn import PixelCNN


def predict_batch(
    x: Union[list, np.ndarray, tf.Tensor],
    model: Union[Callable, keras.Model],
    batch_size: int = int(1e10),
    preprocess_fn: Optional[Callable] = None,
    dtype: Union[Type[np.generic], tf.DType] = np.float32,
) -> Union[np.ndarray, tf.Tensor, tuple, list]:
    """
    Make batch predictions on a model.

    Parameters
    ----------
    x
        Batch of instances.
    model
        tf.keras model or one of the other permitted types defined in Data.
    batch_size
        Batch size used during prediction.
    preprocess_fn
        Optional preprocessing function for each batch.
    dtype
        Model output type, e.g. np.float32 or tf.float32.

    Returns
    -------
    Numpy array, tensorflow tensor or tuples of those with model outputs.
    """
    n = len(x)
    n_minibatch = int(np.ceil(n / batch_size))
    return_np = not isinstance(dtype, tf.DType)
    return_list = False
    preds: Union[list, tuple] = []
    for i in range(n_minibatch):
        istart, istop = i * batch_size, min((i + 1) * batch_size, n)
        x_batch = x[istart:istop]  # type: ignore
        if isinstance(preprocess_fn, Callable):  # type: ignore
            x_batch = preprocess_fn(x_batch)
        preds_tmp = model(x_batch)
        if isinstance(preds_tmp, (list, tuple)):
            if len(preds) == 0:  # init tuple with lists to store predictions
                preds = tuple([] for _ in range(len(preds_tmp)))
                return_list = isinstance(preds_tmp, list)
            for j, p in enumerate(preds_tmp):
                preds[j].append(p if not return_np or isinstance(p, np.ndarray) else p.numpy())
        elif isinstance(preds_tmp, (np.ndarray, tf.Tensor)):
            preds.append(  # type: ignore
                preds_tmp
                if not return_np or isinstance(preds_tmp, np.ndarray)  # type: ignore
                else preds_tmp.numpy()  # type: ignore
            )
        else:
            raise TypeError(
                f"Model output type {type(preds_tmp)} not supported. The model output "
                f"type needs to be one of list, tuple, np.ndarray or tf.Tensor."
            )
    concat = np.concatenate if return_np else tf.concat
    out = cast(
        Union[tuple, tf.Tensor, np.ndarray],
        tuple(concat(p, axis=0) for p in preds) if isinstance(preds, tuple) else concat(preds, axis=0),
    )
    if return_list:
        out = list(out)
    return out


def _get_default_encoder_net(input_shape: Tuple[int, int, int], encoding_dim: int):
    return Sequential(
        [
            InputLayer(input_shape=input_shape),
            Conv2D(64, 4, strides=2, padding="same", activation=relu),
            Conv2D(128, 4, strides=2, padding="same", activation=relu),
            Conv2D(512, 4, strides=2, padding="same", activation=relu),
            Flatten(),
            Dense(encoding_dim),
        ]
    )


def _get_default_decoder_net(input_shape: Tuple[int, int, int], encoding_dim: int):
    return Sequential(
        [
            InputLayer(input_shape=(encoding_dim,)),
            Dense(4 * 4 * 128),
            Reshape(target_shape=(4, 4, 128)),
            Conv2DTranspose(256, 4, strides=2, padding="same", activation=relu),
            Conv2DTranspose(64, 4, strides=2, padding="same", activation=relu),
            Flatten(),
            Dense(math.prod(input_shape)),
            Reshape(target_shape=input_shape),
        ]
    )


def create_model(
    model_type: Union[AE, AEGMM, PixelCNN, VAE, VAEGMM],
    input_shape: Tuple[int, int, int],
    encoding_dim: Optional[int] = None,
    n_gmm: Optional[int] = None,
    gmm_latent_dim: Optional[int] = None,
):
    """
    Create a default model for the specified model type.

    Parameters
    ----------
    model_type
        The model type to create.
    input_shape
        The input shape of the data used.
    encoding_dim
        The target encoding dimensionality.
    n_gmm
        Number of components used in the GMM layer.
    gmm_latent_dim
        Latent dimensionality of the GMM layer.
    """
    input_dim = math.prod(input_shape)
    encoding_dim = int(math.pow(2, int(input_dim.bit_length() * 0.8)) if encoding_dim is None else encoding_dim)
    if model_type == AE:
        return AE(
            _get_default_encoder_net(input_shape, encoding_dim),
            _get_default_decoder_net(input_shape, encoding_dim),
        )

    if model_type == VAE:
        return VAE(
            _get_default_encoder_net(input_shape, encoding_dim),
            _get_default_decoder_net(input_shape, encoding_dim),
            encoding_dim,
        )

    if model_type == AEGMM:
        n_gmm = 2 if n_gmm is None else n_gmm
        gmm_latent_dim = 1 if gmm_latent_dim is None else gmm_latent_dim
        # The outlier detector is an encoder/decoder architecture
        encoder_net = Sequential(
            [
                Flatten(),
                InputLayer(input_shape=(input_dim,)),
                Dense(60, activation=tanh),
                Dense(30, activation=tanh),
                Dense(10, activation=tanh),
                Dense(gmm_latent_dim, activation=None),
            ]
        )
        # Here we define the decoder
        decoder_net = Sequential(
            [
                InputLayer(input_shape=(gmm_latent_dim,)),
                Dense(10, activation=tanh),
                Dense(30, activation=tanh),
                Dense(60, activation=tanh),
                Dense(input_dim, activation=None),
                Reshape(target_shape=input_shape),
            ]
        )
        # GMM autoencoders have a density network too
        gmm_density_net = Sequential(
            [
                InputLayer(input_shape=(gmm_latent_dim + 2,)),
                Dense(10, activation=tanh),
                Dense(n_gmm, activation=softmax),
            ]
        )
        return AEGMM(
            encoder_net=encoder_net,
            decoder_net=decoder_net,
            gmm_density_net=gmm_density_net,
            n_gmm=n_gmm,
        )

    if model_type == VAEGMM:
        n_gmm = 2 if n_gmm is None else n_gmm
        gmm_latent_dim = 2 if gmm_latent_dim is None else gmm_latent_dim
        # The outlier detector is an encoder/decoder architecture
        # Here we define the encoder
        encoder_net = Sequential(
            [
                Flatten(),
                InputLayer(input_shape=(input_dim,)),
                Dense(20, activation=relu),
                Dense(15, activation=relu),
                Dense(7, activation=relu),
            ]
        )
        # Here we define the decoder
        decoder_net = Sequential(
            [
                InputLayer(input_shape=(gmm_latent_dim,)),
                Dense(7, activation=relu),
                Dense(15, activation=relu),
                Dense(20, activation=relu),
                Dense(input_dim, activation=None),
                Reshape(target_shape=input_shape),
            ]
        )
        # GMM autoencoders have a density network too
        gmm_density_net = Sequential(
            [
                InputLayer(input_shape=(gmm_latent_dim + 2,)),
                Dense(10, activation=relu),
                Dense(n_gmm, activation=softmax),
            ]
        )
        return VAEGMM(
            encoder_net=encoder_net,
            decoder_net=decoder_net,
            gmm_density_net=gmm_density_net,
            n_gmm=n_gmm,
            latent_dim=gmm_latent_dim,
        )

    if model_type == PixelCNN:
        return PixelCNN(
            image_shape=input_shape,
            num_resnet=5,
            num_hierarchies=2,
            num_filters=32,
            num_logistic_mix=1,
            receptive_field_dims=(3, 3),
            dropout_p=0.3,
            l2_weight=0.0,
        )

    raise TypeError(f"Unknown model specified: {model_type}.")



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import NamedTuple, Tuple

import numpy as np
import tensorflow as tf


class GaussianMixtureModelParams(NamedTuple):
    """
    phi : tf.Tensor
        Mixture component distribution weights.
    mu : tf.Tensor
        Mixture means.
    cov : tf.Tensor
        Mixture covariance.
    L : tf.Tensor
        Cholesky decomposition of `cov`.
    log_det_cov : tf.Tensor
        Log of the determinant of `cov`.
    """

    phi: tf.Tensor
    mu: tf.Tensor
    cov: tf.Tensor
    L: tf.Tensor
    log_det_cov: tf.Tensor


def gmm_params(z: tf.Tensor, gamma: tf.Tensor) -> GaussianMixtureModelParams:
    """
    Compute parameters of Gaussian Mixture Model.

    Parameters
    ----------
    z : tf.Tensor
        Observations.
    gamma : tf.Tensor
        Mixture probabilities to derive mixture distribution weights from.

    Returns
    -------
    GaussianMixtureModelParams(phi, mu, cov, L, log_det_cov)
        The parameters used to calculate energy.
    """
    # compute gmm parameters phi, mu and cov
    N = gamma.shape[0]  # nb of samples in batch
    sum_gamma = tf.reduce_sum(gamma, 0)  # K
    phi = sum_gamma / N  # K
    mu = tf.reduce_sum(tf.expand_dims(gamma, -1) * tf.expand_dims(z, 1), 0) / tf.expand_dims(
        sum_gamma, -1
    )  # K x D (D = latent_dim)
    z_mu = tf.expand_dims(z, 1) - tf.expand_dims(mu, 0)  # N x K x D
    z_mu_outer = tf.expand_dims(z_mu, -1) * tf.expand_dims(z_mu, -2)  # N x K x D x D
    cov = tf.reduce_sum(tf.expand_dims(tf.expand_dims(gamma, -1), -1) * z_mu_outer, 0) / tf.expand_dims(
        tf.expand_dims(sum_gamma, -1), -1
    )  # K x D x D

    # cholesky decomposition of covariance and determinant derivation
    D = tf.shape(cov)[1]  # type: ignore
    eps = 1e-6
    L = tf.linalg.cholesky(cov + tf.eye(D) * eps)  # K x D x D
    log_det_cov = 2.0 * tf.reduce_sum(tf.math.log(tf.linalg.diag_part(L)), 1)  # K

    return GaussianMixtureModelParams(phi, mu, cov, L, log_det_cov)


def gmm_energy(
    z: tf.Tensor,
    params: GaussianMixtureModelParams,
    return_mean: bool = True,
) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Compute sample energy from Gaussian Mixture Model.

    Parameters
    ----------
    params : GaussianMixtureModelParams
        The gaussian mixture model parameters.
    return_mean : bool, default True
        Take mean across all sample energies in a batch.

    Returns
    -------
    sample_energy
        The sample energy of the GMM.
    cov_diag
        The inverse sum of the diagonal components of the covariance matrix.
    """
    D = tf.shape(params.cov)[1]  # type: ignore
    z_mu = tf.expand_dims(z, 1) - tf.expand_dims(params.mu, 0)  # N x K x D
    z_mu_T = tf.transpose(z_mu, perm=[1, 2, 0])  # K x D x N
    v = tf.linalg.triangular_solve(params.L, z_mu_T, lower=True)  # K x D x D

    # rewrite sample energy in logsumexp format for numerical stability
    logits = tf.math.log(tf.expand_dims(params.phi, -1)) - 0.5 * (
        tf.reduce_sum(tf.square(v), 1)
        + tf.cast(D, tf.float32) * tf.math.log(2.0 * np.pi)  # type: ignore py38
        + tf.expand_dims(params.log_det_cov, -1)
    )  # K x N
    sample_energy = -tf.reduce_logsumexp(logits, axis=0)  # N

    if return_mean:
        sample_energy = tf.reduce_mean(sample_energy)

    # inverse sum of variances
    cov_diag = tf.reduce_sum(tf.divide(1, tf.linalg.diag_part(params.cov)))

    return sample_energy, cov_diag



from daml._internal.models.tensorflow.autoencoder import AE, AEGMM, VAE, VAEGMM, eucl_cosim_features
from daml._internal.models.tensorflow.losses import Elbo, LossGMM
from daml._internal.models.tensorflow.pixelcnn import PixelCNN
from daml._internal.models.tensorflow.utils import create_model

__all__ = ["create_model", "eucl_cosim_features", "AE", "AEGMM", "Elbo", "LossGMM", "PixelCNN", "VAE", "VAEGMM"]



import warnings
from typing import Optional, Tuple

import numpy as np
import scipy


class Parity:
    """
    Class for evaluating statistics of observed and expected class labels, including:

    - Chi Squared test for statistical independence between expected and observed labels

    Parameters
    ----------
    expected_labels : np.ndarray
        List of class labels in the expected dataset
    observed_labels : np.ndarray
        List of class labels in the observed dataset
    num_classes : Optional[int]
        The number of unique classes in the datasets. If this is not specified, it will
        be inferred from the set of unique labels in expected_labels and observed_labels
    """

    def __init__(self, expected_labels: np.ndarray, observed_labels: np.ndarray, num_classes: Optional[int] = None):
        self.set_labels(expected_labels, observed_labels, num_classes)

    def set_labels(self, expected_labels: np.ndarray, observed_labels: np.ndarray, num_classes: Optional[int] = None):
        """
        Calculates the label distributions for expected and observed labels
        and performs validation on the results.

        Parameters
        ----------
        expected_labels : np.ndarray
            List of class labels in the expected dataset
        observed_labels : np.ndarray
            List of class labels in the observed dataset
        num_classes : Optional[int]
            The number of unique classes in the datasets. If this is not specified, it will
            be inferred from the set of unique labels in expected_labels and observed_labels

        Raises
        ------
        ValueError
            If x is empty
        """
        self.num_classes = num_classes

        # Calculate
        observed_dist = self._calculate_label_dist(observed_labels)
        expected_dist = self._calculate_label_dist(expected_labels)

        # Validate
        self._validate_dist(observed_dist, "observed")

        # Normalize
        expected_dist = self._normalize_expected_dist(expected_dist, observed_dist)

        # Validate normalized expected distribution
        self._validate_dist(expected_dist, f"expected for {np.sum(observed_dist)} observations")
        self._validate_class_balance(expected_dist, observed_dist)

        

        self._observed_dist = observed_dist
        self._expected_dist = expected_dist

    def _normalize_expected_dist(self, expected_dist: np.ndarray, observed_dist: np.ndarray) -> np.ndarray:
        exp_sum = np.sum(expected_dist)
        obs_sum = np.sum(observed_dist)

        if exp_sum == 0:
            raise ValueError(
                f"Expected label distribution {expected_dist} is all zeros. "
                "Ensure that Parity.expected_dist is set to a list "
                "with at least one nonzero element"
            )

        # Renormalize expected distribution to have the same total number of labels as the observed dataset
        if exp_sum != obs_sum:
            expected_dist = expected_dist * obs_sum / exp_sum

        return expected_dist

    def _calculate_label_dist(self, labels: np.ndarray) -> np.ndarray:
        """
        Calculate the class frequencies associated with a dataset

        Parameters
        ----------
        labels : np.ndarray
            List of class labels in a dataset

        Returns
        -------
        label_dist : np.ndarray
            Array representing label distributions
        """
        label_dist = np.bincount(labels, minlength=(self.num_classes if self.num_classes else 0))
        return label_dist

    def _validate_class_balance(self, expected_dist: np.ndarray, observed_dist: np.ndarray):
        """
        Check if the numbers of unique classes in the datasets are unequal

        Parameters
        ----------
        expected_dist : np.ndarray
            Array representing expected label distributions
        observed_dist : np.ndarray
            Array representing observed label distributions

        Raises
        ------
        ValueError
            When exp_ld and obs_ld do not have the same number of classes
        """
        exp_n_cls = len(expected_dist)
        obs_n_cls = len(observed_dist)
        if exp_n_cls != obs_n_cls:
            raise ValueError(
                f"Found {obs_n_cls} unique classes in observed label distribution, "
                f"but found {exp_n_cls} unique classes in expected label distribution,"
                "This can happen when some class ids have zero instances in one dataset but "
                "not in the other. When initializing Parity, "
                "try setting the num_classes parameter to the known number of unique class ids, "
                "so that classes with zero instances are still included in the distributions."
            )

    def _validate_dist(self, label_dist: np.ndarray, label_name: str):
        """
        Verifies that the given label distribution has labels and checks if
        any labels have frequencies less than 5.

        Parameters
        ----------
        label_dist : np.ndarray
            Array representing label distributions

        Raises
        ------
        ValueError
            If label_dist is empty
        Warning
            If any elements of label_dist are less than 5
        """
        if not len(label_dist):
            raise ValueError(f"No labels found in the {label_name} dataset")
        if np.any(label_dist < 5):
            warnings.warn(
                f"Labels {np.where(label_dist<5)[0]} in {label_name}"
                " dataset have frequencies less than 5. This may lead"
                " to invalid chi-squared evaluation."
            )
            warnings.warn(
                f"Labels {np.where(label_dist<5)[0]} in {label_name}"
                " dataset have frequencies less than 5. This may lead"
                " to invalid chi-squared evaluation."
            )

    def evaluate(self) -> Tuple[np.float64, np.float64]:
        """
        Perform a one-way chi-squared test between observation frequencies and expected frequencies that
        tests the null hypothesis that the observed data has the expected frequencies.

        This function acts as an interface to the scipy.stats.chisquare method, which is documented at
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chisquare.html
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chisquare.html

        Returns
        -------
        np.float64
            chi-squared value of the test
        np.float64
            p-value of the test
        """
        cs_result = scipy.stats.chisquare(f_obs=self._observed_dist, f_exp=self._expected_dist)

        chisquared = cs_result.statistic
        p_value = cs_result.pvalue
        return chisquared, p_value



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable, Literal, Optional, Tuple

import numpy as np
from scipy.stats import ks_2samp

from .base import BaseUnivariateDrift, UpdateStrategy, preprocess_x


class DriftKS(BaseUnivariateDrift):
    """
    Kolmogorov-Smirnov (K-S) data drift detector with Bonferroni or False Discovery
    Rate (FDR) correction for multivariate data.

    Parameters
    ----------
    x_ref : np.ndarray
        Data used as reference distribution.
    p_val : float, default 0.05
        p-value used for significance of the statistical test for each feature.
        If the FDR correction method is used, this corresponds to the acceptable
        q-value.
    x_ref_preprocessed : bool, default False
        Whether the given reference data `x_ref` has been preprocessed yet. If
        `x_ref_preprocessed=True`, only the test data `x` will be preprocessed at
        prediction time. If `x_ref_preprocessed=False`, the reference data will also
        be preprocessed.
    update_x_ref : Optional[UpdateStrategy], default None
        Reference data can optionally be updated using an UpdateStrategy class. Update
        using the last n instances seen by the detector with
        :py:class:`daml.detectors.LastSeenUpdateStrategy`
        or via reservoir sampling with
        :py:class:`daml.detectors.ReservoirSamplingUpdateStrategy`.
    preprocess_fn : Optional[Callable[[np.ndarray], np.ndarray]], default None
        Function to preprocess the data before computing the data drift metrics.
        Typically a dimensionality reduction technique.
    correction : Literal["bonferroni", "fdr"], default "bonferroni"
        Correction type for multivariate data. Either 'bonferroni' or 'fdr' (False
        Discovery Rate).
    alternative : Literal["two-sided", "less", "greater"], default "two-sided"
        Defines the alternative hypothesis. Options are 'two-sided', 'less' or
        'greater'.
    n_features
        Number of features used in the statistical test. No need to pass it if no
        preprocessing takes place. In case of a preprocessing step, this can also
        be inferred automatically but could be more expensive to compute.
    """

    def __init__(
        self,
        x_ref: np.ndarray,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preprocess_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        correction: Literal["bonferroni", "fdr"] = "bonferroni",
        alternative: Literal["two-sided", "less", "greater"] = "two-sided",
        n_features: Optional[int] = None,
    ) -> None:
        super().__init__(
            x_ref=x_ref,
            p_val=p_val,
            x_ref_preprocessed=x_ref_preprocessed,
            update_x_ref=update_x_ref,
            preprocess_fn=preprocess_fn,
            correction=correction,
            n_features=n_features,
        )

        # Other attributes
        self.alternative = alternative

    @preprocess_x
    def score(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute K-S scores and statistics per feature.

        Parameters
        ----------
        x
            Batch of instances.

        Returns
        -------
        Feature level p-values and K-S statistics.
        """
        x = x.reshape(x.shape[0], -1)
        x_ref = self.x_ref.reshape(self.x_ref.shape[0], -1)
        p_val = np.zeros(self.n_features, dtype=np.float32)
        dist = np.zeros_like(p_val)
        for f in range(self.n_features):
            dist[f], p_val[f] = ks_2samp(x_ref[:, f], x[:, f], alternative=self.alternative, method="exact")
        return p_val, dist



"""
This module contains the implementation of the
FR Test Statistic based estimate and the
KNN based estimate for the Bayes Error Rate

Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4)
https://arxiv.org/abs/1811.06419
"""

from typing import Callable, Dict, Literal, Tuple

import numpy as np
from scipy.sparse import coo_matrix
from scipy.stats import mode

from daml._internal.metrics.base import EvaluateMixin, MethodsMixin

from .utils import compute_neighbors, get_classes_counts, minimum_spanning_tree


def _mst(X: np.ndarray, y: np.ndarray, _: int) -> Tuple[float, float]:
    M, N = get_classes_counts(y)

    tree = coo_matrix(minimum_spanning_tree(X))
    matches = np.sum([y[tree.row[i]] != y[tree.col[i]] for i in range(N - 1)])
    deltas = matches / (2 * N)
    upper = 2 * deltas
    lower = ((M - 1) / (M)) * (1 - max(1 - 2 * ((M) / (M - 1)) * deltas, 0) ** 0.5)
    return upper, lower


def _knn(X: np.ndarray, y: np.ndarray, k: int) -> Tuple[float, float]:
    M, N = get_classes_counts(y)

    # All features belong on second dimension
    X = X.reshape((X.shape[0], -1))
    nn_indices = compute_neighbors(X, X, k=k)
    nn_indices = np.expand_dims(nn_indices, axis=1) if nn_indices.ndim == 1 else nn_indices
    modal_class = mode(y[nn_indices], axis=1, keepdims=True).mode.squeeze()
    upper = float(np.count_nonzero(modal_class - y) / N)
    lower = _knn_lowerbound(upper, M, k)
    return upper, lower


def _knn_lowerbound(value: float, classes: int, k: int) -> float:
    "Several cases for computing the BER lower bound"
    if value <= 1e-10:
        return 0.0

    if classes == 2 and k != 1:
        if k > 5:
            # Property 2 (Devroye, 1981) cited in Snoopy paper, not in snoopy repo
            alpha = 0.3399
            beta = 0.9749
            a_k = alpha * np.sqrt(k) / (k - 3.25) * (1 + beta / (np.sqrt(k - 3)))
            return value / (1 + a_k)
        if k > 2:
            return value / (1 + (1 / np.sqrt(k)))
        # k == 2:
        return value / 2

    return ((classes - 1) / classes) * (1 - np.sqrt(max(0, 1 - ((classes / (classes - 1)) * value))))


_METHODS = Literal["MST", "KNN"]
_FUNCTION = Callable[[np.ndarray, np.ndarray, int], Tuple[float, float]]


class BER(EvaluateMixin, MethodsMixin[_METHODS, _FUNCTION]):
    """
    An estimator for Multi-class Bayes Error Rate using FR or KNN test statistic basis

    Parameters
    ----------
    data : np.ndarray
        Array of images or image embeddings
    labels : np.ndarray
        Array of labels for each image or image embedding
    method : Literal["MST", "KNN"], default "KNN"
        Method to use when estimating the Bayes error rate
    k : int, default 1
        number of nearest neighbors for KNN estimator -- ignored by MST estimator


    See Also
    --------
    `Learning to Bound the Multi-class Bayes Error (Th. 3 and Th. 4) <https://arxiv.org/abs/1811.06419>`_

    """

    def __init__(self, data: np.ndarray, labels: np.ndarray, method: _METHODS = "KNN", k: int = 1) -> None:
        self.data = data
        self.labels = labels
        self.k = k
        self._set_method(method)

    @classmethod
    def _methods(
        cls,
    ) -> Dict[str, _FUNCTION]:
        return {"MST": _mst, "KNN": _knn}

    def evaluate(self) -> Dict[str, float]:
        """
        Calculates the Bayes Error Rate estimate using the provided method

        Returns
        -------
        Dict[str, float]
            ber : float
                The estimated lower bounds of the Bayes Error Rate
            ber_lower : float
                The estimated upper bounds of the Bayes Error Rate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        upper, lower = self._method(self.data, self.labels, self.k)
        return {"ber": upper, "ber_lower": lower}



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

# pyright: reportIncompatibleMethodOverride=false

from typing import Callable, Tuple, cast

import tensorflow as tf
from keras.api._v2 import keras
from keras.api._v2.keras.layers import (
    Dense,
    Flatten,
    Layer,
)


def relative_euclidean_distance(x: tf.Tensor, y: tf.Tensor, eps: float = 1e-12, axis: int = -1) -> tf.Tensor:
    """
    Relative Euclidean distance.

    Parameters
    ----------
    x
        Tensor used in distance computation.
    y
        Tensor used in distance computation.
    eps
        Epsilon added to denominator for numerical stability.
    axis
        Axis used to compute distance.

    Returns
    -------
    Tensor with relative Euclidean distance across specified axis.
    """
    denom = tf.concat(
        [
            tf.reshape(tf.norm(x, ord=2, axis=axis), (-1, 1)),  # type: ignore
            tf.reshape(tf.norm(y, ord=2, axis=axis), (-1, 1)),  # type: ignore
        ],
        axis=1,
    )
    dist = tf.norm(tf.math.subtract(x, y), ord=2, axis=axis) / (tf.reduce_min(denom, axis=axis) + eps)  # type: ignore
    return dist


def eucl_cosim_features(x: tf.Tensor, y: tf.Tensor, max_eucl: float = 1e2) -> tf.Tensor:
    """
    Compute features extracted from the reconstructed instance using the
    relative Euclidean distance and cosine similarity between 2 tensors.

    Parameters
    ----------
    x
        Tensor used in feature computation.
    y
        Tensor used in feature computation.
    max_eucl
        Maximum value to clip relative Euclidean distance by.

    Returns
    -------
    Tensor concatenating the relative Euclidean distance and cosine similarity features.
    """
    if len(x.shape) > 2 or len(y.shape) > 2:
        x = cast(tf.Tensor, Flatten()(x))
        y = cast(tf.Tensor, Flatten()(y))
    rec_cos = tf.reshape(keras.losses.cosine_similarity(y, x, -1), (-1, 1))
    rec_euc = tf.reshape(relative_euclidean_distance(y, x, -1), (-1, 1))
    # rec_euc could become very large so should be clipped
    rec_euc = tf.clip_by_value(rec_euc, 0, max_eucl)
    return cast(tf.Tensor, tf.concat([rec_cos, rec_euc], -1))


class Sampling(Layer):
    """Reparametrization trick. Uses (z_mean, z_log_var) to sample the latent vector z."""

    def call(self, inputs: Tuple[tf.Tensor, tf.Tensor]) -> tf.Tensor:
        """
        Sample z.

        Parameters
        ----------
        inputs
            Tuple with mean and log variance.

        Returns
        -------
        Sampled vector z.
        """
        z_mean, z_log_var = inputs
        batch, dim = tuple(tf.shape(z_mean).numpy().ravel()[:2])  # type: ignore
        epsilon = cast(tf.Tensor, keras.backend.random_normal(shape=(batch, dim)))
        return z_mean + tf.exp(tf.math.multiply(0.5, z_log_var)) * epsilon


class EncoderAE(Layer):
    def __init__(self, encoder_net: keras.Model) -> None:
        """
        Encoder of AE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a keras.Sequential class.
        name
            Name of encoder.
        """
        super().__init__(name="encoder_ae")
        self.encoder_net = encoder_net

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return cast(tf.Tensor, self.encoder_net(x))


class EncoderVAE(Layer):
    def __init__(self, encoder_net: keras.Model, latent_dim: int) -> None:
        """
        Encoder of VAE.

        Parameters
        ----------
        encoder_net
            Layers for the encoder wrapped in a keras.Sequential class.
        latent_dim
            Dimensionality of the latent space.
        name
            Name of encoder.
        """
        super().__init__(name="encoder_vae")
        self.encoder_net = encoder_net
        self.fc_mean = Dense(latent_dim, activation=None)
        self.fc_log_var = Dense(latent_dim, activation=None)
        self.sampling = Sampling()

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        x = cast(tf.Tensor, self.encoder_net(x))
        if len(x.shape) > 2:
            x = cast(tf.Tensor, Flatten()(x))
        z_mean = cast(tf.Tensor, self.fc_mean(x))
        z_log_var = cast(tf.Tensor, self.fc_log_var(x))
        z = cast(tf.Tensor, self.sampling((z_mean, z_log_var)))
        return z_mean, z_log_var, z


class Decoder(Layer):
    def __init__(self, decoder_net: keras.Model) -> None:
        """
        Decoder of AE and VAE.

        Parameters
        ----------
        decoder_net
            Layers for the decoder wrapped in a keras.Sequential class.
        name
            Name of decoder.
        """
        super().__init__(name="decoder")
        self.decoder_net = decoder_net

    def call(self, x: tf.Tensor) -> tf.Tensor:
        return cast(tf.Tensor, self.decoder_net(x))


class AE(keras.Model):
    """
    Combine encoder and decoder in AE.

    Parameters
    ----------
    encoder_net
        Layers for the encoder wrapped in a keras.Sequential class.
    decoder_net
        Layers for the decoder wrapped in a keras.Sequential class.
    """

    def __init__(self, encoder_net: keras.Model, decoder_net: keras.Model) -> None:
        super().__init__(name="ae")
        self.encoder = EncoderAE(encoder_net)
        self.decoder = Decoder(decoder_net)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        z = cast(tf.Tensor, self.encoder(x))
        x_recon = cast(tf.Tensor, self.decoder(z))
        return x_recon


class VAE(keras.Model):
    """
    Combine encoder and decoder in VAE.

    Parameters
    ----------
    encoder_net
        Layers for the encoder wrapped in a keras.Sequential class.
    decoder_net
        Layers for the decoder wrapped in a keras.Sequential class.
    latent_dim
        Dimensionality of the latent space.
    beta
        Beta parameter for KL-divergence loss term.
    """

    def __init__(self, encoder_net: keras.Model, decoder_net: keras.Model, latent_dim: int, beta: float = 1.0) -> None:
        super().__init__(name="vae_model")
        self.encoder = EncoderVAE(encoder_net, latent_dim)
        self.decoder = Decoder(decoder_net)
        self.beta = beta
        self.latent_dim = latent_dim

    def call(self, x: tf.Tensor) -> tf.Tensor:
        z_mean, z_log_var, z = cast(Tuple[tf.Tensor, tf.Tensor, tf.Tensor], self.encoder(x))
        x_recon = self.decoder(z)
        # add KL divergence loss term
        kl_loss = -0.5 * tf.reduce_mean(z_log_var - tf.square(z_mean) - tf.exp(z_log_var) + 1)
        self.add_loss(self.beta * kl_loss)
        return cast(tf.Tensor, x_recon)


class AEGMM(keras.Model):
    """
    Deep Autoencoding Gaussian Mixture Model.

    Parameters
    ----------
    encoder_net
        Layers for the encoder wrapped in a keras.Sequential class.
    decoder_net
        Layers for the decoder wrapped in a keras.Sequential class.
    gmm_density_net
        Layers for the GMM network wrapped in a keras.Sequential class.
    n_gmm
        Number of components in GMM.
    recon_features
        Function to extract features from the reconstructed instance by the decoder.
    """

    def __init__(
        self,
        encoder_net: keras.Model,
        decoder_net: keras.Model,
        gmm_density_net: keras.Model,
        n_gmm: int,
        recon_features: Callable = eucl_cosim_features,
    ) -> None:
        super().__init__("aegmm")
        self.encoder = encoder_net
        self.decoder = decoder_net
        self.gmm_density = gmm_density_net
        self.n_gmm = n_gmm
        self.recon_features = recon_features

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        enc = self.encoder(x)
        x_recon = cast(tf.Tensor, self.decoder(enc))
        recon_features = self.recon_features(x, x_recon)
        z = cast(tf.Tensor, tf.concat([enc, recon_features], -1))
        gamma = cast(tf.Tensor, self.gmm_density(z))
        return x_recon, z, gamma


class VAEGMM(keras.Model):
    """
    Variational Autoencoding Gaussian Mixture Model.

    Parameters
    ----------
    encoder_net
        Layers for the encoder wrapped in a keras.Sequential class.
    decoder_net
        Layers for the decoder wrapped in a keras.Sequential class.
    gmm_density_net
        Layers for the GMM network wrapped in a keras.Sequential class.
    n_gmm
        Number of components in GMM.
    latent_dim
        Dimensionality of the latent space.
    recon_features
        Function to extract features from the reconstructed instance by the decoder.
    beta
        Beta parameter for KL-divergence loss term.
    """

    def __init__(
        self,
        encoder_net: keras.Model,
        decoder_net: keras.Model,
        gmm_density_net: keras.Model,
        n_gmm: int,
        latent_dim: int,
        recon_features: Callable = eucl_cosim_features,
        beta: float = 1.0,
    ) -> None:
        super().__init__(name="vaegmm")
        self.encoder = EncoderVAE(encoder_net, latent_dim)
        self.decoder = decoder_net
        self.gmm_density = gmm_density_net
        self.n_gmm = n_gmm
        self.latent_dim = latent_dim
        self.recon_features = recon_features
        self.beta = beta

    def call(self, x: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        enc_mean, enc_log_var, enc = cast(Tuple[tf.Tensor, tf.Tensor, tf.Tensor], self.encoder(x))
        x_recon = cast(tf.Tensor, self.decoder(enc))
        recon_features = self.recon_features(x, x_recon)
        z = cast(tf.Tensor, tf.concat([enc, recon_features], -1))
        gamma = cast(tf.Tensor, self.gmm_density(z))
        # add KL divergence loss term
        kl_loss = -0.5 * tf.reduce_mean(enc_log_var - tf.square(enc_mean) - tf.exp(enc_log_var) + 1)
        self.add_loss(self.beta * kl_loss)
        return x_recon, z, gamma
